#include "wu_head.h"


/*
 * Functions manipulating the peculiarities of my implementation
 */

IP int_to_maskbits (int mask) {
				/* 
				 * return mask bits in HOST order
				 */
  IP res = 0;
  int i;
  if (mask == WU_IPVX_MAXSIZE)
    return 0;
  if (mask == 0)
    return 0xfFfFfFfF;
  if (WU_IPVX_MAXSIZE < mask  || 0 > mask) 
    wu_error(BUG, "mask %d (in function \"int_to_maskbits()\")", mask);
  i = (WU_IPVX_MAXSIZE - mask);
  while (i--) {
    res = res << 1;
    res++;
  }
  return res;
}

/* inverse function of int_to_maskbits) */
int maskbits_to_int (IP dist) {
	
  int mask = 0;
  while (dist != 0) {
    if (! dist & 1) 
      wu_error (BUG, "called \"maskbits_to_int\" on weird IP");
    dist = dist >> 1;
    mask++;
  }

  mask = WU_IPVX_MAXSIZE - mask;

  if (mask > WU_IPVX_MAXSIZE || mask < 0) 
    wu_error(BUG, "got mask of %d", mask);

  return (mask);
}


int replies_equal (struct wu_reply * a, struct wu_reply * b) {

  if (a->flag == WU_REPLY_TIMEOUT || b->flag == WU_REPLY_TIMEOUT)
    return 0;
  else
    return  (a->respondent.ip_no == b->respondent.ip_no &&
	     a->proto == b->proto &&
	     a->packettype == b->packettype &&
	     a->packetcode == b->packetcode &&
	     a->flag == b->flag
	     );
}


int _replies_equal_ip_and_packttype (struct wu_reply * a, struct wu_reply * b, int mask) {
  IP maskout;
  struct wu_reply aa;
  struct wu_reply bb;

  maskout = 0xFfFfFfFf - int_to_maskbits (mask);
  
  memcpy (&aa, a, sizeof (struct wu_reply));
  memcpy (&bb, b, sizeof (struct wu_reply));
  
  aa.respondent.ip_no &= htonl(maskout); /*  respondent is NET order...? */
  bb.respondent.ip_no &= htonl(maskout);

  return replies_equal (&aa, &bb);
}


int _replies_equal_only_ip (struct wu_reply * a, struct wu_reply *b, int mask) {
  IP maskout = 0xFfFfFfFf - int_to_maskbits (mask);
  IP aa = ntohl (a->respondent.ip_no) & maskout;
  IP bb = ntohl (b->respondent.ip_no) & maskout;
  
  return ( aa == bb ? 1 : 0);
}


int in_iplist (IP  foo, struct iplist * list) {
  if (list == NULL) return 0;
  if (foo == list->ip) return 1;
  return _in_routerlist (foo, list->next);
}


int _in_routerlist (struct wu_reply * foo, struct routerlist * list) {
  if (list == NULL) return 0;
  if (replies_equal (foo, list->reply)) return 1;
  return _in_routerlist (foo, list->next);
}


int in_routerlist (struct routerlist * foo, struct routerlist * list) {
  struct routerlist * bar = foo;
  while (bar != NULL) {
    if (_in_routerlist (bar->reply, list)) return 1;
    bar = bar->next;
  }
  return 0;
}


int _next_farm_range (struct wu_reply * a, struct wu_reply * b) {
  IP a_ho = ntohl(a->respondent.ip_no);
  IP b_ho = ntohl(b->respondent.ip_no);;
  
  if (a_ho + int_to_maskbits(a->farm_mask) == b_ho) return 1;
  return 0;
}


int contiguous_farm_heuristic (struct wu_reply * a,
			       struct wu_reply * b) {
  if (a->flag & WU_FARM && b->flag & WU_FARM) {
    if (_next_farm_range (a, b) || _next_farm_range (b, a))
      return 1;
    else
      return 0;
  }
  else { return 0;  }
}


int replies_equal_heuristic (struct wu_reply * a, 
			     struct wu_reply * b,
			     int mask, 
			     int packettype) {
  
  /* TIMEOUTS should never be equal */
  if (a->flag == WU_REPLY_TIMEOUT || b->flag == WU_REPLY_TIMEOUT)
    return 0;

  /* FARMS should not be equal (checked in contiguous_farm_heuristic()) */
  if (a->flag & WU_FARM || b->flag & WU_FARM)
    return 0;

  if (mask > WU_IPVX_MAXSIZE || mask < 0)
    wu_error(DEBUG_MORE, "replies_equal_heuristic(): mask %d", mask);

  if (packettype == ONLY_CARE_ABOUT_IP) 
    return (_replies_equal_only_ip (a, b, mask));
  else if (packettype == CARE_ABOUT_PACKET_TYPE)
    return (_replies_equal_ip_and_packttype (a, b, mask));
  else 
    wu_error (BUG, "replies_equal... function called wrong");
}


struct wu_reply * get_random_color(int maskin) {
  struct wu_reply * r = (struct wu_reply *) wmalloc (sizeof (struct wu_reply));
  char ip[4];
  int rand;

  get_random_bytes(ip, 4);

  r->respondent = new_ip2(htonl((IP)ip & int_to_maskbits(WU_IPVX_MAXSIZE - maskin)));
  r->flag = WU_REPLY;
  r->proto = IPPROTO_ICMP;
  r->packettype = ICMP_UNREACH;

  rand = get_random_int() % 7;
  switch (rand) {
  case 1:
    r->packetcode = ICMP_UNREACH_NET;
    break;
  case 2:
    r->packetcode = ICMP_UNREACH_HOST_PROHIB;
    break;
  case 3:
    r->packetcode = ICMP_UNREACH_HOST;
    break;
  case 4:
    r->packetcode = ICMP_UNREACH_NET_PROHIB;
    break;
  case 5:
    r->packettype = ICMP_TIMXCEED;
    break;
  default:
    r->packettype = ICMP_SOURCE_QUENCH;
    break;
  }
  return r;
  }


IP split_range (IP l, IP h) {
  /*
   * range is l-h INCLUSIVE 
   * return is HOST order
   */

  if ((h - l) < 0) { wu_error (BUG, "error in split: l > h"); }
  if (h == l) { 
    wu_error (BUG, "possible error in logic: h == l in split"); 
    return l;
  }
  if ( ((h-l) %2) != 1) {wu_error (BUG, "error in programmer's logic: odd range\n"); }
  
  return ( l + ( (h - l + 1) / 2)); /* host order */
}


void * wmalloc (size_t size)  {
  void * res;
  
  res = malloc (size);

  if (res == NULL) 
    wu_error (CRITICAL, "malloc function failed: out of memory");

  return res;
}


int ip_cmp (IP a, IP b) {
  u_long A, B;

  A = /*htonl*/ (a);
  B = /*htonl*/ (b);

  return ( A <= B);
}
