#include "wu_head.h"
#include <stdio.h>

/* EXTERN OPTIONS */
extern char opt_ips;
extern char opt_verbose;
/* EXTERN COUNTER */
extern int numseen;
/* END EXTERN */


struct _stack {
#define MAX_STACK_SIZE 100
  int i;
#define STACK_VERT_LINE_PUSH 1
#define STACK_SPACE 2		
#define STACK_SPACES(x) ((x << 8) | STACK_SPACE)
  int entries[MAX_STACK_SIZE + 1]; 
};
struct _stack the_stack;


void init_stack() { the_stack.i = 0; }


int push (int thing) {
  if (the_stack.i > MAX_STACK_SIZE) {
    wu_error (BUG, "MAX_STACK_SIZE exceeded");    return (-1);  }
  the_stack.entries[the_stack.i++] = thing;

  return (1); 
}

int pop () {
  if (the_stack.i <= 0) {
    wu_error (BUG, "tried pop() from empty stack");
    return (-1); }

  return (the_stack.entries[--the_stack.i]);
}


int colprint (int x) {
  printf ("%c%c%c%c%c%c%c%c", 0x1b, 0x5b, 0x30, 0x31, 0x3b, 0x33, x, 0x6d);
  return 0;
}

#define WU_CYAN 54
#define WU_GREEN 50
#define WU_BOLD 55
#define WU_NORM 59
#define WU_BLU 52 

#define CYAN_ON 	if (! CNTRL.ipdumpmode) colprint(WU_CYAN)
#define BLU_ON 	if (! CNTRL.ipdumpmode) colprint(WU_BLU)
#define BOLD_ON if (! CNTRL.ipdumpmode) colprint(WU_BOLD)
#define NORM_ON if (! CNTRL.ipdumpmode) colprint(WU_NORM)
#define GREEN_ON if (! CNTRL.ipdumpmode) colprint(WU_GREEN)


#ifdef USE_LINES

#define LINES_ON /*GREEN_ON; */	if (! CNTRL.ipdumpmode) printf ("%c%c%c", 27, 4, 14); 
#define LINES_OFF /*NORM_ON; */	if (! CNTRL.ipdumpmode) printf ("%c%c%c", 27, 4, 15);
#define LINES_ON_2 /*GREEN_ON; */ printf ("%c%c%c", 27, 4, 14); 
#define LINES_OFF_2 /*NORM_ON; */ printf ("%c%c%c", 27, 4, 15);


#define STACK_NET_FIRST "k"
#define STACK_NET_MID "u"
#define STACK_NET_LAST "q"
#define STACK_NET_LAST_BRANCH "v"

//these should all be the same length
#define STACK_RTR_FIRST "q+"
#define STACK_RTR_FIRST_BRANCH "w+"
#define STACK_RTR_MID "t+"
#define STACK_RTR_LAST "m+"


#define STACK_HORIZ_LINE "q"
#define STACK_VERT_LINE "~"

#define STACK_INITIAL_BRANCH "t"
#define STACK_INITIAL_LAST_BRANCH "m"

#else  /* using ASCII */

#define STACK_NET_FIRST "-\\"
#define STACK_NET_MID   "-|"
#define STACK_NET_LAST  "--"
#define STACK_NET_LAST_BRANCH "-+"

//these should all be the same length
#define STACK_RTR_FIRST "--"
#define STACK_RTR_FIRST_BRANCH "+-"
#define STACK_RTR_MID   "|-"
#define STACK_RTR_LAST  "`-"

#define STACK_HORIZ_LINE "-"
#define STACK_VERT_LINE "."
#define STACK_INITIAL_BRANCH "|-"
#define STACK_INITIAL_LAST_BRANCH "`-"

#endif


void print_stack() { 
  int i, j;
  int foo; 
  for (i = 0; i < the_stack.i; i++) {
    foo = the_stack.entries[i];
    
    switch (foo & 0x0F) {
    case STACK_VERT_LINE_PUSH:
#ifdef USE_LINES
      LINES_ON_2
#endif
      printf("%s", STACK_VERT_LINE);
#ifdef USE_LINES
      LINES_OFF_2
#endif
      break;
    case STACK_SPACE:
      for (j = 0; j < (foo >> 8); j++)
	printf(" ");
      break;
    default:
      wu_error (BUG, "invalid value on stack");
      break;
    }
  }
}


int _fill_line_ (int f) { 
  while (f--) {
#ifdef USE_LINES
  LINES_ON_2
#endif
    printf ("%s", STACK_HORIZ_LINE); 
#ifdef USE_LINES
  LINES_OFF_2
#endif
    }
  return 0;
}


char * wu_ip_string (IP ip) {
  return libnet_host_lookup (htonl(ip), 0); 
}


int ipswidth (struct iplist * ips) {
  int max = 0; int foo;
  struct iplist * ptr = ips;
  char poo[WU_SMALL_STRING + 1];
  poo[0] = '\0';

  while (ptr != NULL) {
    strcpy (poo, wu_ip_string (ptr->ip));
    foo = strlen (poo);
    if (foo > max) max = foo;
    ptr = ptr->next;
  }

  return max;
}


int netwidth (struct netlist * nets) 
{
  int max = 0; int foo;
  struct netlist * ptr = nets;
  char poo[WU_SMALL_STRING + 1];
  poo[0] = '\0';

  while (ptr != NULL) {
    strcpy (poo, wu_net_string (ptr->host, ptr->mask));
    foo = strlen (poo);
    if (foo > max) max = foo;
    ptr = ptr->next;
  }

  return max;
}


// we want the width of last in list!!!
int routerswidth (struct routerlist * rtrs) 
{
  struct routerlist * ptr = rtrs;
  char poo[WU_SMALL_STRING + 1];

  while (ptr->next != NULL)
    ptr = ptr->next;

  strcpy (poo, wu_router_string (ptr->reply));
  return strlen (poo);
}


char nobranch (struct wu_net * net1, struct wu_net * net2) {
  struct netlist * a, * b;
  char res = 1;
  a = net1->nets; b = net2->nets;
  wu_error (DEBUG_MORE, "nobranch(...)");
  
  if ( a != NULL && b != NULL) {
    while ( a != NULL && b != NULL) {
      if (a->host == b->host && a->mask == b->mask) {
	a = a->next;
	b = b->next;
      }
      else { res = 0; goto jump;}
    }
    if (a != NULL || b != NULL) res = 0;
  }
 jump:
  return res;
}


#define _OPT_NOTRACE (CNTRL.mode & WU_NOTRACE)
int _will_draw_nets () {
  if (_OPT_NOTRACE && opt_ips) return 1;

  if (opt_ips)
    if (CNTRL.mode & WU_SPLIT_MODE)
      if (CNTRL.cur_mask < WU_IPVX_MAXSIZE)
	return 1;

  return 0;
}


//INPUT: wu_net structure N, ...
//OUTPUT: void
void _wu_dfs (struct wu_net * net,  int depth, char drawnets, struct state_vars state) {
#ifndef _OPT_NOTRACE
#define _OPT_NOTRACE (CNTRL.mode & WU_NOTRACE)
#endif 
#define TREEPRINT if (! CNTRL.ipdumpmode )  printf 
#define PRINT_STACK if (! CNTRL.ipdumpmode )  print_stack()
  struct wu_net * NEWNETPTR;
  struct wu_net * NEWNETS = NULL;
  struct netlist * NETLISTPTR;
  struct iplist * IPLISTPTR, * farmptr;
  struct routerlist * RTRPTR;
  IP2 farmip;
  int fieldsize;
  char first;
  char no_reps = 0;
  char willdrawnets;

  char _got_reply = 0;		
  int _dests_reached = 0, _num_replies = 0;
  
  wu_error (DEBUG_MORE, "dfs depth %d from host %s/%d %s", depth,
	    libnet_host_lookup (CNTRL.cur_host, 0),
	    CNTRL.cur_mask,
	    drawnets ? "drawing nets" : "nets same");

  if (drawnets) {
    ///calculate PADDING for nets
    IPLISTPTR = net->reps;
    NETLISTPTR = net->nets;    
    if (IPLISTPTR == NULL) no_reps = 1;

    if (opt_ips && (! no_reps))  fieldsize = ipswidth (IPLISTPTR);
    else fieldsize = netwidth (NETLISTPTR);


    if (opt_ips && (! no_reps)) {
      ///DRAW THE NETS, ITERATE THROUGH IPLIST
      first = 1;
      while (IPLISTPTR != NULL) {

	////DRAW STACK
	if ( depth > 1 || ! first)		
	  PRINT_STACK; 

	////DRAW net + (PADDING - strlen (net))
	TREEPRINT ("%s", wu_ip_string (IPLISTPTR->ip));
	if (! CNTRL.ipdumpmode)
	  _fill_line_ (fieldsize - strlen (wu_ip_string (IPLISTPTR->ip)));
	 fflush(stdout); 

	////DRAW -- if last etc...
#ifdef USE_LINES
	LINES_ON
#endif
	  if (first && IPLISTPTR->next == NULL) { TREEPRINT ("%s", STACK_NET_LAST); }
	  else if (first) {TREEPRINT ("%s\n", STACK_NET_FIRST);}
	  else if (IPLISTPTR->next != NULL) {TREEPRINT ("%s\n", STACK_NET_MID);}

	  else {TREEPRINT ("%s", STACK_NET_LAST_BRANCH);}
	first = 0;  fflush(stdout); 
#ifdef USE_LINES
	LINES_OFF
#endif
	  IPLISTPTR = IPLISTPTR->next;
      }
    } else if ((!opt_ips) || no_reps) {	/* not ips mode */
      ///DRAW THE NETS, ITERATE THROUGH NETLIST
      first = 1;
      while (NETLISTPTR != NULL) {

	////DRAW STACK
	if (depth > 1 || ! first)	
	  PRINT_STACK; 

	////DRAW net + (PADDING - strlen (net))
	if (no_reps) {
	  TREEPRINT ("(%s)", wu_net_string (NETLISTPTR->host, NETLISTPTR->mask));}

	else {
	  TREEPRINT ("%s", wu_net_string (NETLISTPTR->host, NETLISTPTR->mask)); }
	if (! CNTRL.ipdumpmode)
	  _fill_line_ (fieldsize - strlen (wu_net_string (NETLISTPTR->host, NETLISTPTR->mask)));
	 fflush(stdout); 

	////DRAW -- if last etc...
#ifdef USE_LINES
	LINES_ON
#endif
	  if (first && NETLISTPTR->next == NULL) { TREEPRINT ("%s", STACK_NET_LAST); }
	  else if (first) {TREEPRINT ("%s\n", STACK_NET_FIRST);}
	  else if (NETLISTPTR->next != NULL) {TREEPRINT ("%s\n", STACK_NET_MID);}
	  else {TREEPRINT ("%s", STACK_NET_LAST_BRANCH);}
	first = 0;  fflush(stdout); 
#ifdef USE_LINES
	LINES_OFF
#endif
	  NETLISTPTR = NETLISTPTR->next;
      }
    }


    ///PUSH FOR fieldsize of nets
    if (net->next == NULL) {
      push (STACK_SPACES ((strlen (STACK_VERT_LINE)))); }
    else {
      if (depth > 1 || CNTRL.current_ttl < 2 ) {push (STACK_VERT_LINE_PUSH); }
      else { push (STACK_SPACES (( strlen (STACK_VERT_LINE) ))); } }

    push (STACK_SPACES (( (no_reps ? 2 : 0) +
			  fieldsize + 
			  strlen (STACK_NET_LAST) -
			  strlen (STACK_VERT_LINE))));    

    ///DREW THE NETS OR IPS, FINISHED ITERATING THROUGH NETLIST
  } /* drawnets */

  RTRPTR = net->routers;

  ///SET THE LOOP VARS
  _got_reply = 0;
  _dests_reached = 0;
  _num_replies = 0;

  first = 1;			/* reset the little thing */
  while (RTRPTR != NULL) {
    ////DRAW -- if first
    ////DRAW |- if middle
    ////DRAW \- if last
    if (first) {
#ifdef USE_LINES
      LINES_ON
#endif
	if (RTRPTR->next == NULL) { 
	  TREEPRINT ("%s", STACK_RTR_FIRST); 
	}
	else { 
	  TREEPRINT ("%s", STACK_RTR_FIRST_BRANCH); 
	}
#ifdef USE_LINES
      LINES_OFF
#endif
	}
    else if ( RTRPTR->next != NULL) { /* ! first and ! last */
      PRINT_STACK;
#ifdef USE_LINES
      LINES_ON
#endif
	TREEPRINT ("%s", STACK_RTR_MID); 
#ifdef USE_LINES
      LINES_OFF
#endif
	}
    else {			/* last*/
      PRINT_STACK;
#ifdef USE_LINES
      LINES_ON
#endif
	TREEPRINT ("%s", STACK_RTR_LAST); 
#ifdef USE_LINES
      LINES_OFF
#endif
	}
    first = 0;
     fflush(stdout); 

    ////DRAW router
    if (! CNTRL.ipdumpmode ) {
      BOLD_ON;
      TREEPRINT ("%s", wu_router_string (RTRPTR->reply));
      NORM_ON;
      if ( (RTRPTR->reply->flag & WU_FARM) &&
	   opt_ips) {
	farmptr=net->reps;
	TREEPRINT ("\n");
	////push SPACES for FARM PREFIX (lines, etc)
	push (STACK_SPACES (( strlen (STACK_RTR_FIRST)) ));
	PRINT_STACK;
	while (farmptr != NULL) {
	  farmip = new_ip2 (htonl (farmptr->ip /* in ho */)); 
	  BOLD_ON;
	  TREEPRINT ("(%s)", wu_lookup (&farmip));
	  NORM_ON;
	  if (farmptr->next != NULL) { TREEPRINT ("\n"); PRINT_STACK;}
	  farmptr = farmptr->next;
	}
	////pop SPACES for FARM PREFIX
	pop();
      }
    } else if ( CNTRL.ipdumpmode ) {
    }
     
    ///CHECK IF WE CONTINUE SCAN:
  /* wu_dfs checks the wu_reply structure in the wu_net structure if all the pings were timeouts, replys, etc. Here there could be room for improvement in the structure of the program... */
    if (! (RTRPTR->reply->flag & WU_REPLY_TIMEOUT)) _got_reply = 1;
    if  (RTRPTR->reply->flag & WU_FARM) _dests_reached++;
    else if (RTRPTR->reply->pingflag & PING_REPLY) _dests_reached++; 
    _num_replies++;

    if (RTRPTR->next != NULL) TREEPRINT ("\n");
    RTRPTR = RTRPTR->next;
  }
  ///DREW THE ROUTERS, ITERATED THROUGH ROUTER FIELD

  ////push SPACES for ROUTER PREFIX
  push (STACK_SPACES (( strlen (STACK_RTR_FIRST)) ));
  //}

  ///RECURSE
  if (_got_reply == 0) state.timeouts++;
  else state.timeouts = 0;

  if ( (depth < CNTRL.maxdepth) && 
       (state.timeouts < CNTRL.timeout_depth) &&
      (! (_dests_reached == _num_replies)))
       {

    ///FIND THE NEXT NETS
    ////SCAN net, INCREASE TTL
    wu_update_scan ();
    NEWNETS = scan_multi (net);
    ///FOUND THE NEXT NETS
    
    ///NOW FIND IF THE NETS ARE THE SAME
    if (! nobranch (net, NEWNETS)) willdrawnets = 1;
    else willdrawnets = _will_draw_nets (); 
    
    ///IF NOT, PREPARE FOR DRAWING WITH A \n
    if (willdrawnets) {
      TREEPRINT ("\n");
      //      if (! drawnets) pop();
    }
    else if (! willdrawnets) {
      ///PUSH FOR fieldsize of routers
      push (STACK_SPACES ((routerswidth (net->routers)) ));
    } 
    
    ///ITERATE NETS
    NEWNETPTR = NEWNETS;
    while (NEWNETPTR != NULL) {
      _wu_dfs (NEWNETPTR, depth + 1, willdrawnets, state);
      ////ADD EXTRA SPACE
      if (NEWNETPTR->next == NULL && depth > 1 ) 
	{ /*PRINT_STACK; TREEPRINT ("\n");*/ }
      NEWNETPTR = NEWNETPTR->next;
    }
    ///DONE RECURSING NETS
  
    if (! willdrawnets) {
      pop();			/* fieldsize of router */
    } 

    ///RESET TTL
    wu_downdate_scan ();
  } 
  else {
    ///MAKE SURE TO DRAW THE \n since END OF RECURSION
    TREEPRINT ("\n");
  }
  ////DONE RECURSING
  
  pop();			/* router prefix */

  if (drawnets) { 
    pop();			/* line */
    pop();			/* spaces */
  }

  		   /* router line spaces */    
}


/* void  _wu_dfs_new (struct wu_net * net,  int depth, char drawnets) { */
/*   pushed = 0; */

/*   _draw_labels(); */
/*   _draw_nodes(); */
/*   _recurse(); */

/*   _POPALL; */
/* } */


int print_title ( ) {
  //  GREEN_ON;
#ifdef ON_NETWORK
  printf ("(%s at %d hops from %s, %s)\n",
	  wu_net_string (CNTRL.cur_host, CNTRL.cur_mask),
	  CNTRL.current_ttl,
	  libnet_host_lookup (CNTRL.sin.sin_addr.s_addr, 0),
	  CNTRL.device
	  );

#else
  printf ("(pseudo randomly generated map of %s at %d hops)\n",
	  wu_net_string (CNTRL.cur_host, CNTRL.cur_mask),
	  CNTRL.current_ttl
	  );
#endif
  //  NORM_ON;
  return 0;
}
	  

//INPUT: wu_net structure N, ...
//OUTPUT: void
void wu_dfs ( int depth) {
  struct wu_net * NEWNETPTR;
  struct wu_net * NEWNETS = NULL;
  struct netlist foonet;
  struct wu_net foowu;
  struct state_vars state;

  wu_error (DEBUG, "initiating dfs depth %d from host %s/%d", depth,
	    libnet_host_lookup (CNTRL.cur_host, 0),
	    CNTRL.cur_mask);

  if (! CNTRL.ipdumpmode || (opt_verbose))
    print_title ();


  //DO THE FIRST SCAN
  foonet.next = NULL;
  foonet.host = CNTRL.cur_host;
  foonet.mask = CNTRL.cur_mask;
  foowu.nets = &foonet;
  foowu.routers = NULL;
  foowu.reps = NULL;
  foowu.next = NULL;
  state.timeouts = 0;

  NEWNETS = scan_multi (&foowu );

  //FOR EACH n in N:
  NEWNETPTR = NEWNETS;

  //IF ONLY 1 NET (ALL GOES TO SAME PLACE)
  if (nobranch (NEWNETPTR, &foowu)) {
    _wu_dfs (NEWNETPTR, depth + 1, opt_ips, state); 
  }

  //ELSE ITERATE THROUGH NETS
  else {
    while (NEWNETPTR != NULL) {
      if (depth < CNTRL.maxdepth)
	_wu_dfs (NEWNETPTR, depth + 1, 1, state); /* change  */
      NEWNETPTR = NEWNETPTR->next;
    }
  }
}
 

/*
 * give "host/mask" style output
 */
char * wu_net_string (IP ip, int mask) {
  char * foo = (char *) wmalloc (sizeof(char) * WU_SMALL_STRING + 1);

  foo[0] = '\0';
  
  snprintf(foo, WU_SMALL_STRING, "%s/%d",  
 	   libnet_host_lookup(ip, 0),  
 	   mask); 
  return foo; 
}
