#include <libnet.h>
#include <sys/types.h>
#include "wu_head.h"
#include "wu_error.h"


struct _distribution_tree {
  IP ip;
  struct _distribution_tree * l;
  struct _distribution_tree * r;
};


struct _distribution_tree * _free_and_reset_to_hash (struct _distribution_tree * t) {
  if (t == NULL) return NULL;
  _free_and_reset_to_hash (t->l);
  _free_and_reset_to_hash (t->r);
  
  free (t);
  return NULL;
}


#ifdef WU_DEBUG
void dump_random_distribution_dump (struct _distribution_tree* foo, int n) {
  wu_error (DEBUG_MORE, "DISTRIBUTION_TREE (%d): %s", n, libnet_host_lookup (htonl(foo->ip), 0));
  
}


int _dump_random_distribution (struct _distribution_tree * foo, int ln) {
  int count;

  if (foo == NULL) return 0;

  count = _dump_random_distribution (foo->l, ln);
  ln += count;
  dump_random_distribution_dump (foo, ln);
  count += 1;
  ln += 1;
  count += _dump_random_distribution (foo->r, ln);

  return count;

}


void dump_random_distribution (struct _distribution_tree * foo) {
  if (foo == NULL) return;

  _dump_random_distribution (foo, 0);
}
#endif /* debugging code */


int _distribution_tree_insert (IP ip, struct _distribution_tree ** tree) {
  /*
   * return > 0 if IP is already in tree, else insert it and return 0
   */
  wu_error (DEBUG_MORE, "_distribution_tree_insert %s", libnet_host_lookup (ntohl(ip),0));
  if (*tree == NULL) {
    *tree = (struct _distribution_tree *) wmalloc (sizeof (struct _distribution_tree));
    (*tree)->l = NULL;
    (*tree)->r = NULL;
    (*tree)->ip = ip;
    return (1);			
  }

  if ((*tree)->ip == ip) {
    return -1;
  }

  if ((*tree)->ip < ip) 
    return _distribution_tree_insert(ip, &(*tree)->r);
  else
    return _distribution_tree_insert(ip, &(*tree)->l);
}


int _tree_to_array_do (struct _distribution_tree * ran, struct wu_scan ** targets, int lnum, int base_index) {
  struct wu_scan * scan;
  int count = 0;

  wu_error (DEBUG_MORE, "_tree_to_array_do");

  if (ran == NULL) return 0;
  				/* left */
  count += _tree_to_array_do (ran->l, targets, lnum, base_index);
  lnum += count;
				/* middle */
  scan = (struct wu_scan *) wmalloc (sizeof (struct wu_scan));
  scan->target = htonl (ran->ip);
  count++;
  targets[base_index + (lnum++)] = scan;
  targets[base_index + lnum] = NULL;
  				/* right */
  count += _tree_to_array_do (ran->r, targets, lnum, base_index);

  return count;
}


int _tree_to_array (struct _distribution_tree * ran, struct wu_scan ** targets, int base_index) {
  return _tree_to_array_do (ran, targets, 0, base_index);
}


/*
 * random insert into b-tree
 */			
struct _distribution_tree *  _ran_tree_insert (struct _distribution_tree * tree, IP ip) {
  struct _distribution_tree * new;
  
  wu_error (DEBUG_MORE, "_ran_tree_insert");

  if (tree == NULL) {
    new = (struct _distribution_tree *) wmalloc (sizeof (struct _distribution_tree));
    new->l = NULL;
    new->r = NULL;
    new->ip = ip;
    return new;
  }
  else if (rand() & 0x1) {
    tree->l = _ran_tree_insert (tree->l, ip);
    return tree;
  } else {
    tree->r = _ran_tree_insert (tree->r, ip);
    return tree;
  }
}


struct _distribution_tree * _distribution_select_do (IP base_ip_no, int mask, struct _distribution_tree * thedistribution, struct _distribution_tree * ran) {

  IP base_ho = ntohl(base_ip_no);
  IP max_ho = base_ho + int_to_maskbits (mask);  

  if (thedistribution == NULL) { return ran; }

  if (thedistribution->ip >= base_ho && thedistribution->ip <= max_ho) {
    ran = _ran_tree_insert (ran, thedistribution->ip);
    ran = _distribution_select_do (base_ip_no, mask, thedistribution->l, ran);
    ran = _distribution_select_do (base_ip_no, mask, thedistribution->r, ran);
  }
  else if (thedistribution->ip < base_ho) { /* check */
    ran = _distribution_select_do (base_ip_no, mask, thedistribution->r, ran);
  }
  if (thedistribution->ip > max_ho) {
    ran = _distribution_select_do (base_ip_no, mask, thedistribution->l, ran);
  }
  return ran;
}

    
struct _distribution_tree * _distribution_select (IP base_ip_no, int mask, struct _distribution_tree * hash) {
  return _distribution_select_do (base_ip_no, mask, hash,  NULL);
  //check!!
}

 
int _get_current_ips (IP base_ip_ho, 
		      int mask, 
		      struct _distribution_tree * hash, 
		      struct wu_scan ** targets, 
		      int base_index) {
  struct _distribution_tree * subnet;
  int done;

  //  if (hash == NULL) return 0;

  subnet = _distribution_select (htonl(base_ip_ho), mask, hash);
  done = _tree_to_array (subnet, targets, base_index);
  
  return done;
}


struct wu_scan ** _random_distribution_do (struct wu_scan ** targets, struct _distribution_tree ** to_hash, IP base_ip_no, int mask, int index, int num) {

  IP rand;
  int num_ips;			/* from options... */
  int subnet_jump;		/* OPTIONS CHECK */
  IP mask_ip_ho = 0;		/* host order */
  IP high_ip_ho, orig_rand;
  int i, num_tries = 0;
  struct wu_scan * scan;

  /* SANITY CHECKS */  
  if (subnet_jump >= (1 << (WU_IPVX_MAXSIZE - mask))) 
    subnet_jump = 1;
  if (WU_IPVX_MAXSIZE < mask  || 0 > mask) 
    wu_error(BUG, "mask %d out of range in ip choosing\n", mask);

  /* setup */
  high_ip_ho = htonl (base_ip_no) + (1 << (WU_IPVX_MAXSIZE - mask));
  i = (WU_IPVX_MAXSIZE - mask);
  while (i--) {
    mask_ip_ho = mask_ip_ho << 1;
    mask_ip_ho++;
  }
  /*setup done */

  for (i = index; i < num; i++) {
    get_random_bytes((char *)&rand, 4);
    rand &= mask_ip_ho;
    rand += (ntohl(base_ip_no));
    orig_rand = rand;

    num_tries = 0;
    while ( _distribution_tree_insert (rand, to_hash) < 0) {
      num_tries++;
      rand += subnet_jump;    /* helps to avoid clusters */

      if (rand >= high_ip_ho)
	rand -= (1 << (WU_IPVX_MAXSIZE - mask));

      if (rand == orig_rand) {
	rand+=1; 
	if (rand >= high_ip_ho)
	  rand -= (1 << (WU_IPVX_MAXSIZE - mask));
	orig_rand = rand;
      } 

      if (num_tries > (1 << (WU_IPVX_MAXSIZE - mask))) { 
	wu_error(BUG, "network %s/%d only has %d distinct ips, not %d",
		 libnet_host_lookup(base_ip_no, 0),
		 mask,
		 mask_ip_ho + 1,
		 num_ips);
      }
    }
				/* now insert */
    scan = (struct wu_scan *) wmalloc (sizeof (struct wu_scan));
    scan->target = htonl (rand);
    targets[i] = scan;
  }

  targets[i] = NULL;
  return targets;
}

struct wu_scan ** update_probe_targets ( 
					int * numdone) {

#define _MASK ( (CNTRL.mode & WU_MODE_FORCE_BASE) ? WU_IPVX_MAXSIZE : CNTRL.subnet_resolution_mask )
#define _TOT_MASK CNTRL.cur_mask
#define _DENSITY CNTRL.scan_density
#define _BASE CNTRL.cur_host /* this is NET order */

  IP cur_base_ho;
  IP subnet_size;
  IP total_size;
  int numnets, i;
  struct wu_scan ** targets;
  int index = 0, real_density, limit;
  int num_ips;
  
  if (CNTRL.subnet_resolution_mask == 0 && _TOT_MASK == 0) 
    { /* we are "entire internet" */
      numnets = 1;
    }
  else {		
    subnet_size = int_to_maskbits(CNTRL.subnet_resolution_mask) + 1;
    total_size = int_to_maskbits(_TOT_MASK) + 1;
    if (total_size % subnet_size != 0)
      wu_error (BUG, "error in \"update_probe_targets\'");
    numnets = (total_size / subnet_size);
  }

  real_density = 1;
  num_ips = numnets * real_density;

  targets = (struct wu_scan **) wmalloc (sizeof (struct wu_scan *) * num_ips + 1);

  for (i=0; i < numnets; i++) {
    
    index = i * _DENSITY;
    limit = index + real_density;
    cur_base_ho = (i * subnet_size) + ntohl (_BASE);

    index += _get_current_ips (cur_base_ho,
			       _MASK, 
			       CNTRL.targets,
			       targets,
			       index);

    if (index > limit) wu_error (BUG, "error in \"update_probe_targets\"");

    targets = _random_distribution_do (targets, &(CNTRL.targets), 
				 htonl(cur_base_ho), _MASK, 
				 index, limit);
  
  }

  *numdone = num_ips;
  return targets;
}


/* wrappers for drawing ip dump option */
int ip_add_to_the_hash (IP ip) {
  return (_distribution_tree_insert (ip, &CNTRL.routers));
}

int router_add_to_hash ( struct wu_reply * reply) {
  return (_distribution_tree_insert (reply->respondent.ip_no, &CNTRL.routers));
}
