#include "wu_head.h"
#include <stdio.h>


struct _wu_nch_tree {		/* nch == non-contiguous heuristic */
  struct wu_net * wu_net;
  struct _wu_nch_tree * l;
  struct _wu_nch_tree * r;
};


void _free_wu_nch_tree_only_treenodes (struct _wu_nch_tree * t) {
  if (t == NULL) return;
  _free_wu_nch_tree_only_treenodes (t->l);
  _free_wu_nch_tree_only_treenodes (t->r);
  free (t);
}


int _nch_router_cmp (struct routerlist * a, struct routerlist * b, int mask, int replytype) {
   if (a == NULL && b == NULL) return 0;
  if (a== NULL && b != NULL) return -1;
  if (b== NULL && a != NULL) return 1;
  
/*   if ((a->reply->flag &= WU_FARM) && (b->reply->flag &= WU_FARM)) { */
/*     IF WE HAVE FARMS THEY ARE NOT MERGED HERE */
/*     return -1; */
/*   } */
  
  if (contiguous_farm_heuristic (a->reply, b->reply)) return 0;
  if (replies_equal_heuristic (a->reply, b->reply, mask, replytype)) return 0;

  if (a->reply == NULL && b->reply != NULL) return -1;
  if (b->reply == NULL && a->reply != NULL) return 1;
  if (a->reply->flag < b->reply->flag) return -1;
  if (a->reply->flag > b->reply->flag) return 1;
  if (ntohl (a->reply->respondent.ip_no) < ntohl (b->reply->respondent.ip_no)) return -1;
  if (ntohl (b->reply->respondent.ip_no) < ntohl (a->reply->respondent.ip_no)) return -1;

  //should recurse here?
  return 0;
}


struct wu_net * _nch_merge (struct wu_net * a, struct wu_net * b) {
  //merge b INTO a keep a forget about b
  a->routers = routers_insert_list (a->routers, b->routers);
  a->nets = nets_insert_list (a->nets, b->nets);
  a->reps = ips_insert_list (a->reps, b->reps);
  
  return a;
}


struct _wu_nch_tree * _nch_insert (struct wu_net * in, struct _wu_nch_tree * tree, int mask, int replytype, int * res) {

  //res = -1 if in tree

  struct _wu_nch_tree * new;
  int foo;
  int foo2;

  if (tree == NULL) {
    new = (struct _wu_nch_tree *) wmalloc (sizeof (struct _wu_nch_tree));
    new->l = NULL;
    new->r = NULL;
    new->wu_net = in;

    *res = 0;
    return new;
  }

  foo = _nch_router_cmp (in->routers, tree->wu_net->routers, mask, replytype);

  switch (foo) {
  case (-1):
    tree->l = _nch_insert (in, tree->l, mask, replytype, &foo2);
    *res = foo2;
    break;
  case (1):
    tree->r = _nch_insert (in, tree->r, mask, replytype, &foo2);
    *res = foo2;
    break;
  default:
    tree->wu_net = _nch_merge (tree->wu_net, in);
    
    *res = -1;
    break;
  }

  return tree;
}



/*
 * ASSUME THE "HEURISTIC" ALGORITHMS MODIFY THEIR ARGUMENTS
 */
struct wu_net * wu_noncontig_heuristic (struct wu_net * in, 
					int mask, 
					int replytype) {

  struct wu_net * ptr;
  struct wu_net * prev;
  struct wu_net * tmp;
  struct _wu_nch_tree * tree = NULL;
  int res;

  ptr = in;
  tree = _nch_insert (ptr, tree, mask, replytype, &res);

  prev = ptr;
  ptr = ptr->next;

  while (ptr != NULL) {
    tree = _nch_insert (ptr, tree, mask, replytype, &res);

    if (res != 0) {
      prev->next = ptr->next;
      tmp = ptr;
      ptr = ptr->next;
      free (tmp);
    } else {
      prev = ptr;
      ptr = ptr->next;
    }
  }

  _free_wu_nch_tree_only_treenodes (tree);

  return in;
}


struct wu_net * _w_a_h
(struct wu_net * in, 
 int mask, 
 int replytype,
 struct wu_net * last) {
  
  if (in == NULL) return NULL; 

  /*
   * check if the last one is the same as this one
   */
  if (last->routers->reply == NULL || in->routers->reply == NULL) 
    wu_error (BUG, "programmer: bad list struct in _w_a_h");

  if (replies_equal_heuristic (last->routers->reply,
			       in->routers->reply,
			       mask, 
			       replytype) ||
      contiguous_farm_heuristic (last->routers->reply, in->routers->reply)
      ) {
    /*
     * if they are equal
     */
    last->nets = nets_insert_list(in->nets, last->nets);
    last->routers = routers_insert_list(in->routers, last->routers);
    last->reps = ips_insert_list(in->reps, last->reps);

    /*
     * skip "in"
     */
    free (in);

    return _w_a_h (in->next, mask, replytype, last);
  }

  else {
    in->next = _w_a_h (in->next, mask, replytype, in);
    return in;
  }
}


struct wu_net * wu_adjacent_heuristic
(struct wu_net * in, 
 int mask, 
 int replytype)
{
  /*
   * modifies its argument!
   */

  struct wu_net * first;
  if (in == NULL) return NULL;

  first = in;

  in->next = _w_a_h (in->next, mask, replytype, first);
  return in;
}


/*
 * (n^2 algorithm)
 */
#define _ECMP_MAX_NETS 100
struct _ecmp_array {
  struct wu_net * net;
  int netnum;
};


/*
 * ecmp == equal cost multi path				
 */
struct wu_net * wu_ecmp_heuristic
(struct wu_net * in, 
 int mask,
 int replytype) {
  
  struct _ecmp_array * testarray[_ECMP_MAX_NETS + 1];
  int index = 0;
  int array_size;
  int back_index;
  int back_netnum;
  int cur_netnum;
  struct wu_net * wu_net_ptr;
  struct wu_net * res;		

  wu_net_ptr = in; 
  index = 0;
  while (wu_net_ptr != NULL) {
    testarray[index] = (struct _ecmp_array *)wmalloc (sizeof (struct _ecmp_array));
    testarray[index]->net = wu_net_ptr;
    testarray[index]->netnum = index;
    index++;
    if (index > _ECMP_MAX_NETS)
      wu_error (BUG, "Too many networks to analyse in e.c.m.p. algorithm\n");
    wu_net_ptr = wu_net_ptr->next;
  }
  
  array_size = index;		/* index is 0 ... arraysize - 1 */
  
  for (index = 0; index < array_size; index++) {
    for (back_index = 0; back_index < index; back_index++) {
      if (replies_equal_heuristic (testarray[index]->net->routers->reply,
				   testarray[back_index]->net->routers->reply,
				   mask, 
				   replytype)) {
	back_netnum = testarray[back_index]->netnum;
	while (back_index <= index) {
	  testarray[back_index]->netnum = back_netnum;
	  back_index++;	
	}
      }
    }
  }

  wu_error (DEBUG_MORE, "ECMP networks:");
  for (index = 0; index < array_size; index++) 
    wu_error (DEBUG_MORE, "%d ", testarray[index]->netnum);
  
  cur_netnum = -666;
  for (index = 0; index < array_size; index++) {

    if (testarray[index]->netnum != cur_netnum) {
      cur_netnum = testarray[index]->netnum;
      if (index == 0) {
	wu_net_ptr = (struct wu_net *) wmalloc (sizeof (struct wu_net));
	res = wu_net_ptr;
      }
      else {
	wu_net_ptr->next = (struct wu_net *) wmalloc (sizeof (struct wu_net));
	wu_net_ptr = wu_net_ptr->next;
      }
      wu_net_ptr->routers = NULL;
      wu_net_ptr->reps = NULL;
      wu_net_ptr->nets = NULL;
      wu_net_ptr->next = NULL;
    }     
    
    wu_net_ptr->nets = 
      nets_insert_list (testarray[index]->net->nets, wu_net_ptr->nets);
    wu_net_ptr->routers = 
      routers_insert_list (testarray[index]->net->routers, wu_net_ptr->routers);
    wu_net_ptr->reps = 
      ips_insert_list ( testarray[index]->net->reps, wu_net_ptr->reps);
    
    free (testarray[index]->net);
    free (testarray[index]);
  }
  
  return res;
}
