#include "wu_head.h"
#include <stdio.h>


struct wu_net * node2list (WU_NODE * node, struct wu_net ** in) {
  /*
   * return head of list, update in to point to tail of list
   */
  struct wu_net * head;
  struct wu_net * tmpmiddle;
  struct wu_net * tmptail;
  struct wu_net * newnet;

  if (node == NULL) wu_error (BUG, "node2list(): null WU_NODE node");

  if (node->flag == FLAG_AMAL || node->flag == FLAG_UNREP) {
    newnet = (struct wu_net *) wmalloc (sizeof (struct wu_net));
    
    newnet->nets = NULL;
    newnet->routers = NULL;
    newnet->reps = NULL;
    newnet->next = NULL;

    newnet->nets = nets_insert (node->range_l, node->range_h, newnet->nets);
    newnet->routers = routers_insert_list (node->color, newnet->routers);
    if (node->reps != NULL)
      newnet->reps = node->reps;

    if (in != NULL)
      *in = newnet;
    return newnet;
  }
  
  else if (node->flag == FLAG_OPEN) {
    /*
     * left and right cannot both be null
     */
    if (node->left == NULL) {
      head = node2list (node->right, &tmptail); }
    else if (node->right == NULL) {
      head = node2list (node->left, &tmptail); }
    else {
      head = node2list (node->left, &tmpmiddle); 
      tmpmiddle->next = node2list (node->right, &tmptail); }
 
    if (in != NULL)
      *in = tmptail;
    return head;
  }
  
  else if (node->flag == FLAG_UNREP) {
    wu_error(BUG, "\"node2list\", bad node given");
  }

  else {
    wu_error(BUG, "\"node2list\", unknown node flag");
  }

  exit (WEX_NO);
}


void debug_wu_net (struct wu_net * sf) {
				/* 
				 * traverse a wu_net list struct 
				 */

  if (sf == NULL) return;
  wu_error (DEBUG_MORE, "catting list...");
  printf ("networks:\t");
  printf ("%s\n", wu_nets_string (sf->nets));
  fflush(stdout);

  printf ("routed by:\t");
  printf ("%s\n", wu_routers_string (sf->routers));
  fflush(stdout);

  printf ("ips tested:\t");
  printf ("%s\n", wu_ips_string (sf->reps));
  fflush(stdout);

  printf ("\n");

  debug_wu_net (sf->next);

  return;
}


void traverse_wu_net (struct wu_net * sf) {
				/* 
				 * traverse a wu_net list struct 
				 */
  
  if (sf == NULL) return;

  printf ("net [0]: ");
  printf ("%s/%d\n", libnet_host_lookup(sf->nets->host, 0), sf->nets->mask);
  
  printf ("reply [0]: ");
  printf ("%s\n", wu_router_string (sf->routers->reply));

  printf ("\n");

  traverse_wu_net (sf->next);

  return;
}
