#include <libnet.h>
#include <pcap.h>
#include "wu_error.h"
#include <stdio.h>
#include <unistd.h>
#include "wu_head.h"
#include <sys/time.h>
#include <stdlib.h>
#include <string.h>

#define WU_MAIN

#ifdef WU_DEBUG
char wu_debug = 1;
#else
char wu_debug = 0;
#endif


/* OPTIONS HERE */
char opt_tracert = 0;
char opt_dns = 0;
char opt_ips = 0;
char opt_farms = 0;
#define WU_DEFAULT_HEURISTIC 0
char opt_heuristic = WU_DEFAULT_HEURISTIC;
char opt_verbose = 1;


/* COUNTERS HERE */
int packsent = 0;
int packrec = 0;
int bytesent = 0;
int byterec = 0;
int numseen = 0;
struct timeval start_time;
struct timeval end_time;


/* GLOBAL CONTROL STRUCT HERE */
struct wu_control CNTRL;


void _usage (char * foo) {
  printf ("\n%s\t---\t a depth first map of a network\n\n", foo);
  printf ("usage:\n"
	  "\t%s <HOSTNAME> <MASK> <HOPS> [-a | -e | -H] [OPTIONS]\n" 
//	  "\t%s -l <HOSTNAME> <FIRST-HOP> <MAX-HOP> [OPTIONS]\n" 
	  "\n"
	  "\tMASK\t\t0-%d\n"
	  "\tHOPS\t\thops away to start map\n"
	  "\t[-r <depth>]\trecursion depth of search\n"
	  "\t[-d <mask>]\tinitial subnet search size\n"
	  "\t[-P <depth>]\tdepth to extend search past timeouts\n"
	  //	  "\t[-x <int>]\tprobe density parameter\n"
	  "\t[-i <iface>]\tinterface to use (eg: eth0, ppp0 ...)\n" 
	  "\t[-t <timeout>]\ttimeout on probes in ms\n"
	  "\t[-c <count>]\tnumber of probes to sent to each address (max %d)\n"
	  "\t[-N]\t\tdo dns lookups\n"
	  "\t[-I]\t\tonly display ips, not routing table guesses\n"
	  "\t[-T]\t\tdon't force traceroute - pick new ips at each depth\n"
	  "\t[-b]\t\tpick only lowest ip in each subnet for test\n"
//	  "\t[-l]\t\ttraceroute-like mode\n"
	  "\t[-f]\t\tidentify server or dialup farms\n"
	  "\t[-a]\t\tuse adjacent network heuristic\n"
	  "\t[-e]\t\tuse equal cost multi path heuristic\n"
	  "\t[-n]\t\tuse non-contiguous IP routing space heuristic\n"
	  "\t[-m <mask>]\tmask to use for comparing routers\n"
	  "\t[-s]\t\tdump list of unique ips/hostnames to stdout (don't draw map)\n"
	  "\t[-V]\t\tbe as non verbose as possible\n"
	  "\t[-D]\t\tprint debugging output to stderr\n"
	  "\n",
//	  foo,
	  foo,
	  WU_IPVX_MAXSIZE,
	  WU_MAX_QUERIES
	  );
}

 
int more_usage (char * foo) {
  printf ("%s ... [ -a | -e | -H] ...\n",
	  foo);
  exit (WEX_OK);
}


int wu_wrapup () {
  int time_ms, in_bw, out_bw, in_pack_bw, out_pack_bw, tot_bw, tot_pack_bw;

  if (opt_verbose) {
    time_ms =
      ((end_time.tv_sec - start_time.tv_sec) * 1000) +
      ((end_time.tv_usec - start_time.tv_usec) / 1000);
	    
    in_bw = byterec * 1000 / time_ms; 
    out_bw = bytesent * 1000 / time_ms;
    tot_bw = (bytesent  + byterec) * 1000 / time_ms;
    in_pack_bw = (packrec * 1000) / time_ms;
    out_pack_bw = (packsent * 1000) / time_ms;
    tot_pack_bw = ((packrec + packsent) * 1000) / time_ms;

    printf ("\n");

    printf ("sent %d packets / %d bytes, rec'd %d packets / %d bytes from %d ips in %d ms\n", packsent, bytesent, packrec, byterec, numseen, time_ms);
    printf ("bandwidth: %d bytes/sec (%d in  %d out), %d packets/sec (%d in  %d out)\n", 
	    tot_bw,
	    in_bw,
	    out_bw,
	    tot_pack_bw,
	    in_pack_bw,
	    out_pack_bw);

  }
  return 0;
}


int wu_init_random() {
				/* PSEUDO random */
  struct timeval tv;
  gettimeofday (&tv, NULL);

  srand ((tv.tv_sec ^ tv.tv_usec) ^ 0xbadbeef);
  
  return 0;
}


#define ARGS_NEEDED_TR_MODE 3
#define ARGS_NEEDED_NORM 3
#define TRACERT_MODE (opt_tracert)
int main (int argc, char ** argv) {

  char arg;
  IP host_ho, mask_ho;
  char _got_optd = 0;

  wu_init_control ();
  wu_init_random();


  while ( (arg = getopt(argc, argv, "Aabc:d:Defh:Ii:lm:nNpP:r:sS:Tt:Vx:")) != EOF) {
    switch (arg) {
    case 'a':
      if (opt_heuristic != WU_DEFAULT_HEURISTIC) 
	wu_error (WARNING, "more than one heuristic option specified, using \"-a\"");
      opt_heuristic = WU_ADJACENT;
      break;
    case 'b':
      CNTRL.mode |= WU_MODE_FORCE_BASE;
      break;
    case 'c':
      CNTRL.opt_probe_cnt = atoi (optarg);
      if (CNTRL.opt_probe_cnt > WU_MAX_QUERIES) {
	CNTRL.opt_probe_cnt = WU_MAX_QUERIES;
	wu_error (WARNING, "setting query number to maximum allowed of %d", 
		  WU_MAX_QUERIES);}
      break;
    case 'd':
      CNTRL.subnet_resolution_mask = atoi (optarg);
      _got_optd++;
      break;
    case 'l':
      opt_tracert++;
      break;
    case 'h':
      _usage (argv[0]); exit (WEX_OK);
      break;
    case 'T':
      CNTRL.mode |= WU_NOTRACE;
      break;
    case 'm':
      CNTRL.sim_mask = atoi (optarg);
      break;
    case 'P':
      CNTRL.timeout_depth = atoi (optarg);
      if (CNTRL.timeout_depth < 1) wu_error (FATAL, "-P option: timeouts must be >= 1");
      break;
    case 'p':
      CNTRL.packtype = 1;
      break;
    case 'I':
      opt_ips++;
      break;
    case 'N':
      opt_dns++;
      break;
    case 's':
      CNTRL.ipdumpmode = 1;
      break;
     case 'x': 
       CNTRL.scan_density = atoi (optarg); 
       break; 
    case 'i':
      CNTRL.device = (char *)wmalloc(sizeof(char) * sizeof(optarg) + 1);
      strcpy(CNTRL.device, optarg);
      break;
    case 'V':
      opt_verbose = 0;
      break;
    case 't':
      CNTRL.opt_reading_wait.tv_sec = atoi(optarg) / 1000;
      CNTRL.opt_reading_wait.tv_usec = (atoi(optarg) * 1000) % 1000000;
      break;
    case 'D':
      wu_debug++;
      break;
    case 'r':
      if (TRACERT_MODE) 
	wu_error (WARNING, "-r is invalid option in this mode");
      CNTRL.maxdepth = atoi(optarg);
      break;
    case 'e':
      if (opt_heuristic != WU_DEFAULT_HEURISTIC) 
	wu_error (WARNING, "more than one heuristic option specified, using \"-e\"");
      opt_heuristic = WU_ECMP;
      break;
    case 'n':
      if (opt_heuristic != WU_DEFAULT_HEURISTIC) 
	wu_error (WARNING, "more than one heuristic option specified, using \"-n\"");
      opt_heuristic = WU_NONCONTIG;
      break;
    case 'f':
      opt_farms = 1;
      break;
    case 'A':
      break;
    default:
      _usage(argv[0]); exit (WEX_NO);
      break;
    }
  }

  if (TRACERT_MODE) {	/* trace */
    if (optind + ARGS_NEEDED_TR_MODE > argc) {  _usage(argv[0]); exit (WEX_NO); }
    if (optind + ARGS_NEEDED_TR_MODE < argc) { _usage (argv[0]); exit (WEX_NO);}
    CNTRL.cur_host = libnet_name_resolve (argv[optind], 1);
    CNTRL.current_ttl = atoi(argv[optind + 1]);
    CNTRL.maxdepth = atoi (argv[optind + 2]) - atoi (argv[optind + 1]) + 1;
    CNTRL.cur_mask = WU_IPVX_MAXSIZE;
    CNTRL.subnet_resolution_mask = WU_IPVX_MAXSIZE;
    if (CNTRL.maxdepth < 0) {
      wu_error(WARNING, "refusing to do backward tracert: assuming you meant from %d to %d hops", atoi (argv[optind + 2]), atoi (argv[optind + 1]));
      CNTRL.current_ttl = atoi(argv[optind + 2]);
      CNTRL.maxdepth = atoi (argv[optind + 1]) - atoi (argv[optind + 2]) + 1;
    }
    if (CNTRL.maxdepth == 0)  _usage (argv[0]), exit (WEX_NO);
    if (_got_optd)
      wu_error(WARNING, "-d option not valid with traceroute-like mode");
  }
  else {
    if (optind + ARGS_NEEDED_NORM > argc ) { _usage (argv[0]); exit (WEX_NO);    }
    if (optind + ARGS_NEEDED_NORM < argc) { _usage (argv[0]); exit (WEX_NO);}
    CNTRL.cur_host = libnet_name_resolve (argv[optind], 1);
    CNTRL.cur_mask = atoi (argv[optind + 1]);
    CNTRL.current_ttl = atoi(argv[optind + 2]);
  }

  //some sanity checks for mask
  if (CNTRL.cur_mask < 0 || CNTRL.cur_mask > WU_IPVX_MAXSIZE) 
    wu_error (FATAL, "netmask must be between 0 and %d", WU_IPVX_MAXSIZE);

  if (! TRACERT_MODE) {
  //check subnet_resolution_mask
    if (! _got_optd) {
      if (CNTRL.cur_mask < WU_IPVX_MAXSIZE) {
	CNTRL.subnet_resolution_mask = CNTRL.cur_mask ;}
      else {
	CNTRL.subnet_resolution_mask = WU_IPVX_MAXSIZE; } }
    if (CNTRL.subnet_resolution_mask < CNTRL.cur_mask)
      wu_error (FATAL, "-d option: can't divide a classless \"%d\" subnet into \"%d\" subnets",  CNTRL.subnet_resolution_mask, CNTRL.cur_mask);
    if (CNTRL.subnet_resolution_mask > WU_IPVX_MAXSIZE ||
	     CNTRL.subnet_resolution_mask < 0    )
      wu_error (FATAL, "-d option: no such thing as a mask of %d in IPv4", CNTRL.subnet_resolution_mask);
  }

  //make sure that the base host makes sense
  if ((CNTRL.cur_host) == -1) {	/* PROBably name_resolve failed... */
    wu_error (DEBUG_MORE, "host lookup function returned 255.255.255.255");
    wu_error (FATAL, "host %s not found!", argv[optind]); }
  host_ho = ntohl(CNTRL.cur_host);
  mask_ho = int_to_maskbits(CNTRL.cur_mask);
  if (host_ho & mask_ho) {
    host_ho &= (0xffffffff - mask_ho);
    CNTRL.cur_host = htonl(host_ho);
    if (opt_verbose) 
      wu_error(DEBUG, "setting base host to %s  (%s & %s)",
	       libnet_host_lookup (CNTRL.cur_host, 0),
	       argv[optind],
	       libnet_host_lookup (-1 - htonl (int_to_maskbits (atoi(argv[optind + 1]))), 0));
  }

  //check for incompatible options
  if (CNTRL.mode & WU_MODE_FORCE_BASE && CNTRL.mode & WU_NOTRACE)
    wu_error (WARNING, "-T option overrides -b option");

  //check ttl option independent of mode
  if (CNTRL.current_ttl <= 0)
    wu_error (WARNING, "you set initial ttl at %d. it should probably be > 0.", CNTRL.current_ttl);

  //finish setting up the control struct
  CNTRL.base_host = CNTRL.cur_host;
  CNTRL.base_mask = CNTRL.cur_mask;

#ifdef ON_NETWORK
  if (wu_init_network() < 0)
    wu_error(FATAL, "couldn't initialize network. interface %s should be up and %s may need suid permission, or specify interface with -i <INTERFACE>.", 
	     CNTRL.device,
	     argv[0]
	     );

  wu_error(DEBUG, "initialized network");
#else
  wu_error(DEBUG, "Not using network. Generating pseudo random map.");
#endif

  gettimeofday (&start_time, NULL);

  if (TRACERT_MODE)
    wu_tracert ();
  else 
    wu_dfs (  0);

  gettimeofday (&end_time, NULL);

#ifdef ON_NETWORK
  wu_close_network();
#endif

  wu_wrapup ();

  wu_shutdown();

  return 0;
}
