#include "wu_error.h"
#include <libnet.h>
#include <pcap.h>
#include "wu_head.h"

extern int packsent;
extern int bytesent;


#ifdef ON_NETWORK
int
wu_init_network()
{
  char errbuf[LIBNET_ERR_BUF + 1]; 

				/* libnet man page wrong */
  if (libnet_select_device(&CNTRL.sin, (char**)&CNTRL.device, errbuf) == -1)
    {
      return (-1);
    }
  
  /*
   *  Open the raw socket
   */
  CNTRL.sock = libnet_open_raw_sock(IPPROTO_RAW);
  if (CNTRL.sock < 0)    {        return (-1);       }
  
  /*
   *  Open the pcap packet capturing interface.
   */
  CNTRL.pcap = pcap_open_live(CNTRL.device, WU_SNAPLEN, 0, PCAP_READ_TIMEOUT, errbuf);
  if (!(CNTRL.pcap))    {        return (-1);    }

  if ( (CNTRL.linkoff = get_datalink_offset(CNTRL.pcap)) < 0) {
    wu_error(WARNING, "couldn't figure out your linklayer offset");
    return (-1);
  }

  return (1);
}
#endif /* ON_NETWORK */


#ifdef ON_NETWORK
int wu_close_network () {
  close (CNTRL.sock);
  pcap_close (CNTRL.pcap);
  return (1);
}
#endif /* ON_NETWORK */


int
wu_packet_probe( 
		      struct wu_scan * scan)
{
  struct timeval tv;
  int /*i, c,*/ packetsize, finalpacketsize;

  if ( (scan->packet = (u_char *)wmalloc (WU_MAX_PACKET_SIZE)) == NULL) 
    wu_error(CRITICAL, "could not initialize packet, out of memory!!!");

  /*
   *  Build the transport header and payload portion of the packet.
   */
  switch (CNTRL.protocol)
    {
    case IPPROTO_UDP:
      packetsize = wu_packet_build_udp( scan);
      break;
    default:
      wu_error(BUG, "Unimplemented transport protocol\n");
      break;
    }


#ifdef ON_NETWORK
  finalpacketsize = libnet_write_ip (CNTRL.sock, scan->packet, packetsize);
#else
  finalpacketsize = packetsize;
#endif

  if (finalpacketsize < 0) {
    wu_error(CRITICAL, "Couldn't write packet\n");
    return(-1);
  }
  else
    if (finalpacketsize !=  packetsize) {
      wu_error(CRITICAL, "Only wrote %d bytes to network", finalpacketsize);
    }


  /* 
   * Update scan
   */
  packsent++;
  bytesent += finalpacketsize;
  gettimeofday(&tv, NULL);  
  scan->timesent = tv;
  scan->flag = WU_SENT_MULTI;
  //  scan->numsent++;
  return (1);
}


int
wu_packet_build_udp( struct wu_scan * scan)
{
#define UDP_DATALEN 12	

  char udpdata[UDP_DATALEN];
  get_random_bytes(udpdata, UDP_DATALEN);

  /*
   *  Build the IP header portion, debug later
   */
  libnet_build_ip(LIBNET_UDP_H + UDP_DATALEN,
		  0, /* IP type of service */
		  getpid(), /* IP id */
		  IP_DF, /* IP fragmentation bits */
		  CNTRL.current_ttl, /* IP time to live */
		  IPPROTO_UDP, /* transport protocol */
		  CNTRL.sin.sin_addr.s_addr, 
		  (scan->target), /* *** destination IP */
		  NULL,                                   /* IP payload */
		  0,                                      /* IP payload size */
		  (u_char *)scan->packet);            /* packet memory */
  
  if (
      libnet_build_udp(get_random_ushort(),		   /* source UDP port */
		       get_random_ushort(),		   /* dest UDP port */
		       (u_char *) udpdata,                           /* payload ... */
		       UDP_DATALEN,
		       (u_char *)scan->packet + LIBNET_IP_H
		       ) < 0)
    wu_error(CRITICAL, "could not construct UDP packet");

  /* packet memory */
  if (libnet_do_checksum((u_char*) scan->packet, 
			 IPPROTO_UDP,
			 LIBNET_UDP_H + UDP_DATALEN ) < 0) 
    wu_error(CRITICAL, "could not perform UDP checksum");
  
  return (LIBNET_IP_H + LIBNET_UDP_H + UDP_DATALEN);
}


#ifdef ON_NETWORK
int				/* source: firewalk, Packet Factory */
get_datalink_offset(pcap_t *pd)
{
  int offset_len;

  switch (pcap_datalink(pd))
    {
    case DLT_SLIP:
      offset_len = 0x10;
      break;
    case DLT_RAW:
      offset_len = 0x0;
      break;             
    case DLT_PPP:
      offset_len = 0x04;
      break;
    case DLT_EN10MB:
    default:
      offset_len = ETH_H;     /* default to ethernet */
      break;
    }
  return (offset_len);
}
#endif /* on the network */
