#include <libnet.h>
#include <stdio.h>
#include <string.h>
#include "wu_error.h"
#include "wu_head.h"

/*
 * strings and functions for output
 */

/*
 * imitate traceroute and add new output 
 */
struct wu_exotic_out {
  int flag;
  char * string; };

struct wu_exotic_out icmpunreachstrings [] = { 
    { ICMP_UNREACH_NET, "!N" },
    { ICMP_UNREACH_HOST, "!H" },
    { ICMP_UNREACH_PROTOCOL, "!P" },
    { ICMP_UNREACH_PORT, "!p" },
    { ICMP_UNREACH_NEEDFRAG, "!F" },
    { ICMP_UNREACH_SRCFAIL, "!S" },
    { ICMP_UNREACH_ISOLATED, "!I" },
    { ICMP_UNREACH_NET_UNKNOWN, "?N" },
    { ICMP_UNREACH_HOST_UNKNOWN, "?H" },
    { ICMP_UNREACH_NET_PROHIB, "!A" },
    { ICMP_UNREACH_HOST_PROHIB, "!C" },
    { ICMP_UNREACH_FILTER_PROHIB, "|F" }, 
    { ICMP_UNREACH_TOSNET, "%N" },
    { ICMP_UNREACH_TOSHOST, "%H" },
    { ICMP_UNREACH_HOST_PRECEDENCE, "#" },
    { ICMP_UNREACH_PRECEDENCE_CUTOFF, "#/" },
    { 0, NULL }
};
  

struct wu_exotic_out icmptimxceedstrings [] = {
  { ICMP_TIMXCEED_REASS, "<R" }, 
  { ICMP_TIMXCEED_INTRANS, "<+" },
  { 0, NULL }
};


struct wu_exotic_out icmpredirectstrings[] = {
  { ICMP_REDIRECT_NET, "@N" },
  { ICMP_REDIRECT_HOST, "@H" },
  { ICMP_REDIRECT_TOSNET, "&N" },
  { ICMP_REDIRECT_TOSHOST, "&H" },
  { 0, NULL }
};


struct wu_exotic_out icmpparamprobstrings[] = {
  { ICMP_PARAMPROB_OPTABSENT, "=" },
  { 0, NULL } 
};


struct wu_exotic_out icmpsourcequenchstrings[] = {
  { ICMP_SOURCEQUENCH, "$" },
  { 0, NULL }
};


#define WU_TTL_PROBLEM_STRING "^"
#define WU_ICMP_PROTOCOL_VIOLATION_STRING "~"


/* should be done with pointers but... */
char * stringlookup (int flag, struct wu_exotic_out strings[]) {
  int i = 0; 

  while (strings[i].string != NULL) {
    if (strings[i].flag == flag) 
      return (strings[i].string);
    i++;
  }
  
  return ("");
}


char * wu_router_string (struct wu_reply * router) {
  char * buffer = (char *) wmalloc (sizeof (char) * WU_BIG_STRING + 1);

  buffer[0] = '\0';
  
  if (router == NULL) {
    wu_error(BUG, "in wu_output, tree struct not amalgamated");
    return ("unknown");
  }
  
  if (router->flag == WU_REPLY_TIMEOUT) {
    sprintf (buffer, "%s", "*"); 
  }
  else if (router->flag &= WU_FARM) {
    wu_lookup(&(router->respondent));
    
  if ( strcmp (libnet_host_lookup (router->respondent.ip_no, 0), 
		router->respondent.hostname)) 
    sprintf (buffer, "%s/%d farm at %s", 
	     libnet_host_lookup (router->respondent.ip_no, 0),
	     router->farm_mask,
	     router->respondent.hostname);
  else 
    sprintf (buffer, "%s/%d farm",
	     router->respondent.hostname,
	     router->farm_mask
	     );
  }
  else if (
	   (! (router->flag & (WU_TTL_PROBLEM | WU_ICMP_PROTOCOL_VIOLATION)))
	   && router->proto == IPPROTO_ICMP
	   && router->packettype == ICMP_TIMXCEED
	   && router->packetcode == ICMP_TIMXCEED_INTRANS )  {
    sprintf (buffer, "%s", wu_lookup_2(&(router->respondent)));
  }
  else {
    sprintf (buffer, "%s", wu_lookup_2(&(router->respondent)));

    //    strcat (buffer, " (");
    strcat (buffer, " ");

    if (router->flag & WU_TTL_PROBLEM) 
      strcat (buffer, WU_TTL_PROBLEM_STRING);
    if (router->flag & WU_ICMP_PROTOCOL_VIOLATION) 
      strcat (buffer, WU_ICMP_PROTOCOL_VIOLATION_STRING);
 
    switch (router->proto) {
    case (IPPROTO_ICMP):
      switch (router->packettype) {
      case (ICMP_UNREACH):
	strcat (buffer, stringlookup(router->packetcode, icmpunreachstrings));
	break;
      case (ICMP_TIMXCEED):
	strcat (buffer, stringlookup (router->packetcode, icmptimxceedstrings));
	break;
      case (ICMP_REDIRECT):
	strcat (buffer, stringlookup  (router->packetcode, icmpredirectstrings));
	break;
      case (ICMP_PARAMPROB):
	strcat (buffer, stringlookup (router->packetcode, icmpparamprobstrings));
	break;
      case (ICMP_SOURCEQUENCH):	/* no code */
	strcat (buffer, stringlookup (router->packettype, icmpsourcequenchstrings));
	break;
      default:
	strcat (buffer, "unknown icmp"); // %d %d", router->packettype, router->packetcode);
	break;
      }
      break;
    default:
      strcat (buffer, "non-icmp protocol"); // %d", router->proto);
      break;
    }
    //    strcat (buffer, ")");
    strcat (buffer, "");
  }
  return (buffer);
}


char * wu_nets_string (struct netlist * nets) {
  char * netstring = (char *) wmalloc (sizeof(char) * WU_BIG_STRING + 1);
  char foo[WU_SMALL_STRING + 1];
  struct netlist * netptr = NULL;
  netstring[0] = '\0';
  
  if (nets == NULL) {
    strcat (netstring, "none");
    return (char *)netstring;
  }

  else {
    netptr = nets;
    while (netptr != NULL) {
      foo[0] = '\0';
      snprintf(foo, WU_SMALL_STRING, "%s/%d", 
	       libnet_host_lookup(netptr->host, 0), 
	       netptr->mask);
      strcat (netstring, foo);
      if (netptr->next != NULL)
	strcat (netstring, " ");
      netptr=netptr->next;
    }
    return netstring;
  }
}


char * wu_ips_string (struct iplist * ips) {
  char * ipstring = (char *) wmalloc (sizeof(char) * WU_BIG_STRING + 1);
  char foo[WU_SMALL_STRING + 1];
  struct iplist * ipptr = NULL;

  ipstring[0] = '\0';
  
  if (ips == NULL) {
    strcat (ipstring, "none");
    return (char *)ipstring;
  }

  else {
    ipptr = ips;
    while (ipptr != NULL) {
      foo[0] = '\0';
      snprintf(foo, WU_SMALL_STRING,
	       "%s", 
	       libnet_host_lookup(htonl(ipptr->ip), 0));
      strcat (ipstring, foo);
      if (ipptr->next != NULL)
	strcat (ipstring, ", ");
      ipptr=ipptr->next;
    }
    return (char *) ipstring;
  }
}

char * wu_routers_string (struct routerlist * routers) {
  char * routerstring = (char *) wmalloc (sizeof(char) * WU_BIG_STRING + 1);
  char foo[WU_BIG_STRING + 1];
  struct routerlist * routerptr = NULL;
  routerstring[0] = '\0';
  
  if (routers == NULL) {
    strcat (routerstring, "none");
    return (char *)routerstring;
  }

  else {
    routerptr = routers;
    while (routerptr != NULL) {
      foo[0] = '\0';
      sprintf(foo, "%s", 
	      wu_router_string(routerptr->reply));
      strcat (routerstring, foo);
      if (routerptr->next != NULL)
	strcat (routerstring, "  ");
      routerptr=routerptr->next;
    }
    return (char *) routerstring;
  }
}
