#include <libnet.h>
#include "wu_error.h"
#include "wu_head.h"

extern char opt_tracert;

#ifndef ON_NETWORK
extern char gen_farm_mode;
#endif

/*
 * check a packet against one sent in a scan
 * update the status and info of the scan if necessary
 * return > 0 if there is some reply to the scan
 */

#ifdef ON_NETWORK
int
wu_packet_verify( struct wu_scan * scan, char * packet)
{
  /* structs to help dig info from packets */
  struct libnet_ip_hdr* respip, * sentip;
  struct libnet_tcp_hdr* resptcp;
  struct libnet_udp_hdr* respudp;
  struct libnet_icmp_hdr* respicmp;
  struct libnet_ip_hdr* respicmpip;
  struct libnet_udp_hdr* respicmpudp;
  struct libnet_tcp_hdr* respicmptcp;
  struct wu_reply* reply;
  struct timeval tv;

#define _ALLOCATE_REPLY reply=(struct wu_reply *) wmalloc (sizeof (struct wu_reply))
  /* 
   * check that there is a packet and it could be for us
   */
  if (scan->flag == WU_SENT_MULTI && scan->packet != NULL) {
    /*
     * check that the packet is to our ip
     */
    sentip = (struct libnet_ip_hdr *) scan->packet;
    respip = (struct libnet_ip_hdr*) (packet + CNTRL.linkoff);
    if (respip->ip_dst.s_addr == CNTRL.sin.sin_addr.s_addr) {
      /*
       * now check the protocol
       */
      switch (respip->ip_p) { 
	/*
	 * an ICMP pakit detects ok
	 */
      case IPPROTO_ICMP:
	/*
	 * check what type of ICMP 
	 */
	respicmp = (struct libnet_icmp_hdr*) ((u_char*)respip  + LIBNET_IP_H);
	switch (respicmp->icmp_type) {
	  /*
	   * a timxceed in which case the original packet is returned
	   */
	case ICMP_UNREACH:
	case ICMP_TIMXCEED:
	  respicmpip = 
	    (struct libnet_ip_hdr *) ((u_char *) respip + LIBNET_IP_H + 8);
	  /*
	   * check that the dst of orig was that of our scan
	   */
	  if (respicmpip->ip_dst.s_addr == sentip->ip_dst.s_addr) { 
	    /*
	     * we accept this as an answer to probe, and we build the reply struct
	     */
	    scan->flag = WU_GOT_MULTI;
	    _ALLOCATE_REPLY;
	    reply->respondent = new_ip2(respip->ip_src.s_addr);
	    reply->flag = WU_REPLY;
	    reply->proto = IPPROTO_ICMP;
	    reply->packettype = respicmp->icmp_type;
	    reply->packetcode = respicmp->icmp_code;
	    gettimeofday (&tv, NULL);
	    reply->pingtime = wu_tv_diff_ms (tv, scan->timesent);
	    if (respip->ip_src.s_addr == sentip->ip_dst.s_addr)
	      reply->pingflag = PING_REPLY;
	    else 
	      reply->pingflag = NO_PING;
	    /*
	     * check for weird behaviors
	     */
	    /* this is broken */
	    //	      if (memcmp(respicmpip, sentip, LIBNET_IP_H) != 0) {
	    //		wu_error(DEBUG_MORE, "icmp protocol violation or bad programming?");
	    //		reply->flag |= WU_ICMP_PROTOCOL_VIOLATION; }
	    if (respip->ip_ttl <= 1)
	      reply->flag |= WU_TTL_PROBLEM;
	    /*
	     * good, got a reply and exit
	     */
	    wu_error(DEBUG_MORE, "got reply");

	    wu_update_replies (scan, reply);

	    //need to output right away for tracert opt
	    if (opt_tracert) 
 	      if (! CNTRL.ipdumpmode) {
		printf ("%s ", wu_router_string (reply));
		fflush (stdout);
	      }

	    return (1);
	  }
	  break;
	default:
	  break;
	} /* end switch icmp type */
      default:
	break;
      } /* end switch packet protocol */
    }
  }
    
  return (0);			/* didn't get interesting packet */
}


#else
int wu_bogus_reply (struct wu_scan * scan) {
  /* structs to help dig info from packets */
  struct wu_reply* reply;
#define _ALLOCATE_REPLY reply=(struct wu_reply *) wmalloc (sizeof (struct wu_reply))

  scan->flag = WU_GOT_MULTI;
  _ALLOCATE_REPLY;
  reply->respondent = gen_farm_mode ? new_ip2(scan->target) : new_ip2(0 + rand());
  reply->flag = WU_REPLY;
  reply->proto = IPPROTO_ICMP;
  reply->packettype = ICMP_TIMXCEED;
  reply->packetcode = 0;
  reply->pingtime = abs(rand()) % 5000;
  reply->pingflag = gen_farm_mode ? PING_REPLY : NO_PING;

  if (rand() & 0x1) 
    if (rand() & 0x1)
      reply->flag = WU_REPLY_TIMEOUT;


  if (! gen_farm_mode) {
  if (rand() & 0x1) {
    if (rand() & 0x1)
    reply->flag = WU_REPLY_TIMEOUT;
  }
 else {
    if (rand() & 0x1) {
      if (rand() & 0x1) 
	reply->respondent = new_ip2 (10);
      else
	reply->respondent = new_ip2 (69);
    }
    else {
      if (rand() & 0x1) 
	reply->respondent = new_ip2 (666);
      else
	reply->respondent = new_ip2 (555);
    }
  }
  } 


  wu_update_replies (scan, reply);
  return 0x666;
}
#endif /* ON THE NETWORK */


/*
 * simple aray insert
 */
void wu_update_replies (struct wu_scan * scan, struct wu_reply * reply) {
  int n = scan->numreplies;

  if (n > WU_MAX_QUERIES) 
    wu_error(BUG, "more replies than MAX_QUERIES)");

  scan->replies[n] = reply;
  scan->replies[n+1] = NULL;
  scan->numreplies++;

  return;
}
