#include <libnet.h>
#include <pcap.h>
#include "wu_error.h"
#include "wu_head.h"

extern char opt_tracert, opt_dns;
extern struct wu_control CNTRL;
extern int packrec, byterec;
extern int numseen;

#ifndef ON_NETWORK
char gen_farm_mode = 0;
#endif


int _avg_pingtime (struct wu_scan * x) {
  struct wu_reply * ptr;
  int avg = 0;
  int num = 0;
  int i_and_i_rasta = 0;

  ptr = x->replies[i_and_i_rasta];

  while (ptr != NULL) {
    if (ptr->flag != WU_REPLY_TIMEOUT) {
      avg += ptr->pingtime;
      num ++;
    }
    ptr = x->replies[++i_and_i_rasta];
  }

  if (num > 0) return (avg / num);
  return -1;
}


int _tracert_continue_print (struct wu_scan * r) {
  int index = r->numreplies - 1;
#define _TR_CONT_OUT printf ("%s ", wu_router_string (r->replies[index]))

  if (index == 0) 
    _TR_CONT_OUT;
  else if (! replies_equal (r->replies[index], r->replies[index - 1]))
    _TR_CONT_OUT;
  
  if (r->replies[index]->flag != WU_REPLY_TIMEOUT)
    printf ("%d ms  ", r->replies[index]->pingtime);
  else printf (" ");
}


WU_NODE * wu_scan () {
  struct wu_scan ** targetselection;
  struct wu_reply * reply_timeout = (struct wu_reply *)wmalloc (sizeof (struct wu_reply));
  int i, j, lo_window, hi_window, numtoscan;
  int avg; 
  struct wu_scan * currenttarget;
  struct timeval tv;
  WU_NODE * wu_subnets = NULL;

#ifdef ON_NETWORK
  u_char * packet;
  struct pcap_pkthdr pc_hdr;
#endif

#ifndef ON_NETWORK
  gen_farm_mode = /*rand() & *//*(rand() & (*/rand()  & 0x1/* ));*/;
#endif

    /* { 0, WU_TIMEOUT, 0, 0 , 0 } */
  reply_timeout->respondent = new_ip2 (0);
  reply_timeout->flag = WU_REPLY_TIMEOUT; 
  reply_timeout->proto = 0;
  reply_timeout->packettype = 0;
  reply_timeout->packetcode = 0;
  reply_timeout->pingtime = -1;
  reply_timeout->pingflag = NO_PING;


  targetselection = wu_select_targets ( &numtoscan);

  wu_error(DEBUG, "probing %d addresses on the %s/%d subnet at %d hops",
	   numtoscan,
 	   libnet_host_lookup(CNTRL.cur_host, 0),
	   CNTRL.cur_mask,
	   CNTRL.current_ttl);

  /*
   * initialize defaults in struct
   */
  i=0;
  while (targetselection[i] != NULL) {
    targetselection[i]->flag = WU_NOT_PROBED;
    targetselection[i]->numreplies = 0;
    targetselection[i]->replies[0] = NULL;
    targetselection[i]->packet = NULL;
    i++;
  }

  lo_window = 0;			
  hi_window = CNTRL.opt_parprobes;	
  if (hi_window > /*numtoscan*/ i ) hi_window = i; 
  

				/*
				 *  ###### MAIN SCAN LOOP ######
				 */
#define _MORE_PROBES_TO_DO (currenttarget->numreplies < CNTRL.opt_probe_cnt)

#define _PROBE_INCOMPLETE  (_MORE_PROBES_TO_DO &&			\
			    (currenttarget->flag == WU_GOT_MULTI || 	\
			     currenttarget->flag == WU_NOT_PROBED ) )

#define _LOOP_BEGIN     for (i = lo_window; i < hi_window; i++ ) { \
    currenttarget=targetselection[i];
#define _LOOP_END }



  while (lo_window < hi_window) {


    /* SEND PROBES NOT SENT */
    _LOOP_BEGIN			
      if (_PROBE_INCOMPLETE)
	{
	  wu_packet_probe( currenttarget); 
	  wu_error(DEBUG, "%s:\tsent probe", libnet_host_lookup(currenttarget->target, 0));
	}
    _LOOP_END


      /* READ REPLIES TO PROBES */
#ifdef ON_NETWORK
      for (; packet = (u_char *)pcap_next( CNTRL.pcap, &pc_hdr); ) {
	_LOOP_BEGIN
	  if (wu_packet_verify ( currenttarget, packet) > 0) {

	    packrec++;	   
	    byterec += pc_hdr.caplen; // ??? ->len 

	    wu_error(DEBUG, "%s:\treplied", 
		     libnet_host_lookup(currenttarget->target, 0));

	    if (CNTRL.ipdumpmode) {
	      if ( ip_add_to_the_hash (currenttarget->target ) > 0) {
		printf ("%s\n", libnet_host_lookup (currenttarget->target, opt_dns));
		numseen++;
	      }
	    }
	    else if ( ip_add_to_the_hash (currenttarget->target ) > 0) { numseen++; }

	    if (targetselection[hi_window] != NULL && (! _MORE_PROBES_TO_DO))
	      hi_window++;
	  }
	_LOOP_END
	  }
#else  /* ! ON_NETWORK */
    _LOOP_BEGIN
      if (currenttarget->flag == WU_SENT_MULTI) {
	wu_bogus_reply (currenttarget);
	if (opt_tracert && currenttarget->flag == WU_GOT_MULTI) 
	  if (! CNTRL.ipdumpmode) {
	    //output needed right away
	    _tracert_continue_print (currenttarget);
	    fflush (stdout);
	  }
      }
    if (targetselection[hi_window] != NULL && (! _MORE_PROBES_TO_DO)) 
      hi_window++; 
    _LOOP_END
#endif // ON_NETWORK 


      /* CHECK FOR TIMEOUT  */
      gettimeofday(&tv, NULL);
    _LOOP_BEGIN			
      if (currenttarget->flag == WU_SENT_MULTI) {
	if ((tv.tv_sec >= 
	     currenttarget->timesent.tv_sec + CNTRL.opt_reading_wait.tv_sec 
	     &&
	     tv.tv_usec > 
	     currenttarget->timesent.tv_usec + CNTRL.opt_reading_wait.tv_usec) 
	    ||
	    (tv.tv_sec > 
	     currenttarget->timesent.tv_sec + CNTRL.opt_reading_wait.tv_sec)) {
	  wu_error(DEBUG, "%s:\tprobe timed out", 
		   libnet_host_lookup(currenttarget->target, 0));

	  currenttarget->flag = WU_GOT_MULTI; 
	  wu_update_replies (currenttarget, reply_timeout);

	  //need to output right away for tracert opt
	  if (opt_tracert) 
	    if (! CNTRL.ipdumpmode) {
	      _tracert_continue_print (currenttarget);
	      fflush (stdout);
	    }
	}

	if (tv.tv_sec < currenttarget->timesent.tv_sec) 
	  wu_error(CRITICAL, "clock skew detected");
      }
    _LOOP_END


      /* COMPLETE IF ALL PINGS ARE DONE */
      _LOOP_BEGIN		
      if (! _MORE_PROBES_TO_DO) {
	if ( CNTRL.opt_probe_cnt > 1) 
	  wu_error (DEBUG, "%s: finished all %d probes", 
		    libnet_host_lookup (currenttarget->target, 0),
		    CNTRL.opt_probe_cnt);
	currenttarget->flag = WU_MULTI_DONE;
      }
    _LOOP_END


      /* RE ADJUST WINDOW */
      _LOOP_BEGIN		
      if (currenttarget->flag == WU_MULTI_DONE) {
	if (i == lo_window) lo_window++;	
	if (currenttarget->packet != NULL) free((u_char *)currenttarget->packet); 
	if (targetselection[hi_window] != NULL) hi_window++; 
      }
    _LOOP_END
  
      }    
				/*
				 * #### MAIN SCAN LOOP DONE ####
				 */


				/* 
				 * NOW ANALYSIS 
				 */

#define TR_PROBES_TIME_OUT ""

  /*
   * The main Wu algorithm:
   */

  for (i = 0; (currenttarget = targetselection[i]) != NULL; i++) {
    //need to output right away for tracert opt here: should only have 1 targetc
    if (! CNTRL.ipdumpmode && opt_tracert) {
      avg = _avg_pingtime (currenttarget);
      if (avg >= 0 && CNTRL.opt_probe_cnt > 1)
	printf ("(%d ms avg)", avg);
      else
	printf ("%s", TR_PROBES_TIME_OUT);
    };
    
    for (j = 0; j < CNTRL.opt_probe_cnt; j++) {
      wu_subnets = wu_insert (ntohl (CNTRL.cur_host), 
			      ntohl (CNTRL.cur_host) + 
			      int_to_maskbits (CNTRL.cur_mask), 
			      ntohl (currenttarget->target), 
			      currenttarget->replies[j], 
			      wu_subnets);
    }
  }

  /*
   * Clean up the Wu tree:
   */
  amalgamate (wu_subnets);

  /*
   * should free up the target array too
   */
  return (wu_subnets);
}
