#include "wu_head.h"


void traverse (WU_NODE * tree, int mask, int maskin) {
                     
  struct iplist * listptr;
  if (tree == NULL) {
    wu_error (BUG, "error in tree struct: NULL tree\n");
  }

  else if (tree->flag == FLAG_OPEN) {
    traverse (tree->left, mask + 1, maskin);
    traverse (tree->right, mask + 1, maskin);
  }

  else if ( (tree->flag == FLAG_AMAL) || (tree->flag == FLAG_UNREP)) {
    printf ( "%s/%d\t->%s\t", 
	     libnet_host_lookup (ntohl (tree->range_l), 0),
	     mask,
	     wu_routers_string(tree->color));
    if (tree->flag == FLAG_UNREP) { printf ("( none "); }
    else {
      listptr = tree->reps;
      printf("( ");
      while (listptr != NULL) {
	printf ("%s ", 
		libnet_host_lookup (ntohl (listptr->ip), 0));
	listptr = listptr->next;
      }

    }
    printf(")\n");
  }
  else { wu_error (BUG, "error in tree struct: unknown flag"); }
}


void amalgamate (WU_NODE * tree) {
  if (tree == NULL) {
    wu_error(DEBUG_MORE, "empty tree");
    return;
  }
  if (tree->flag == FLAG_UNREP) {
    wu_error (BUG, "empty tree wrong flag UNREP in amalgamate() programmers fault");
    return;
  }
  rmd (tree);			/* righmost defined */
  lmd (tree);			/* leftmost defined */
}


struct routerlist * lmd (WU_NODE * tree) {
  struct routerlist * res;

  if ( (tree->right == NULL && tree->left != NULL) ||
       (tree->left != NULL && tree->right == NULL)) {
    wu_error(BUG, "lmd() encountered a bad unbalanced tree structure, aborting");

  }
  if (tree->left->flag == tree->right->flag == FLAG_UNREP) {
    wu_error(BUG, "lmd() encountered a bad tree structure, aborting");

  }

  switch (tree->flag) {
  case FLAG_AMAL:
    return (tree->color);
    break;
  case FLAG_OPEN:
    if (tree->left->flag & (FLAG_AMAL | FLAG_OPEN)) {
      if (tree->right->flag & (FLAG_AMAL | FLAG_OPEN))
	lmd (tree->right);
      return (lmd (tree->left)); }
    else if (tree->left->flag == FLAG_UNREP) {
      res = lmd (tree->right);
      tree->left->color = routers_insert_list (res, tree->left->color);
      return (res);
    }
    break;
  default:
    wu_error(BUG, "bad tree flag encountered by lmd(), aborting");

    break;
  }

  return (0x0);
}


struct routerlist * rmd (WU_NODE * tree) {		
  struct routerlist * res;

  if ( (tree->left == NULL && tree->right != NULL) ||
       (tree->right != NULL && tree->left == NULL)) {
    wu_error(BUG, "rmd() encountered a bad unbalanced tree structure, aborting");

  }
  if (tree->right->flag == tree->left->flag == FLAG_UNREP) {
    wu_error(BUG, "rmd() encountered a bad tree structure, aborting\n");

  }

  switch (tree->flag) {
  case FLAG_AMAL:
    return (tree->color);
    break;
  case FLAG_OPEN:
    if (tree->right->flag & (FLAG_AMAL | FLAG_OPEN)) {
      if (tree->left->flag & (FLAG_AMAL | FLAG_OPEN))
	rmd (tree->left);
      return (rmd (tree->right));}
    else if (tree->right->flag == FLAG_UNREP) {
      res = rmd (tree->left);
      tree->right->color = routers_insert_list (res, tree->right->color);
      return (res);
    }
    break;
  default:
    wu_error(BUG,"bad tree flag encountered by rmd(), aborting");

    break;
  }

  return 0x0;
}    


WU_NODE * _wu_insert (
		  IP range_l,	/* range INCLUDE range_l */
		  IP range_h,	/* range INCLUDES range_h */
		  IP rep,	/* rep is HOST order */
		  struct routerlist * color, 
		  WU_NODE * tree) {

  WU_NODE * new;
  IP split;
  IP foo;
  struct iplist * listptr;

  
				/* libnet_host_lookup returns static buf!! */
  wu_error (DEBUG_MORE, "_wu_insert: %s in %s/%d",
	    libnet_host_lookup (htonl(rep),0),
	    libnet_host_lookup (htonl(range_l),0),
	    maskbits_to_int (range_h - range_l));

  if (tree == NULL) {           /* insert at leaf */
    if ( (range_l > rep ) || (range_h < rep)) {
      wu_error (BUG,"rep out of range %s/%d on _wu_insert (...,NULL)",
		libnet_host_lookup (htonl (range_l), 0),
		maskbits_to_int (range_h -         range_l));
    }
    new = (WU_NODE *) wmalloc (sizeof (WU_NODE));
    new->left = NULL;
    new->right = NULL;
    new->flag = FLAG_AMAL;
    new->color = routers_insert_list (color, NULL);
    new->range_l = range_l;
    new->range_h = range_h;
    new->reps = NULL;
    new->reps = ips_insert (rep, new->reps);
    return (new);
  }
  else {                       /* tree not null */
    if ( (tree->range_l > rep) || (tree->range_h < rep)) {
      wu_error (BUG,"_wu_insert: rep out of range %s/%d",
		libnet_host_lookup (htonl (tree->range_l), 0),
		maskbits_to_int (tree->range_h - tree->range_l));
    }
    
    if ( tree->flag == FLAG_UNREP ) {
      tree->reps = ips_insert (rep, tree->reps);
      tree->flag = FLAG_AMAL;
      tree->color = routers_insert_list (color, tree->color);
      return (tree); }

    else  if (tree->flag == FLAG_OPEN) { /* open insert */
      if ( (range_l != tree->range_l) || (range_h != tree->range_h))
	wu_error (BUG,"tree structure, ranges out in insert, open node\n");
      split = split_range (range_l, range_h);

      if (rep < split) {		/* rep in range_l */
	tree->left = _wu_insert (range_l, split - 1, rep, color, tree->left); }
      else {			/* rep in range_h */
	tree->right = _wu_insert (split, range_h, rep, color, tree->right); }
    } /* end open insert */

    else if (tree->flag == FLAG_AMAL) {	/* amal insert */
      if ( (range_l != tree->range_l) || (range_h != tree->range_h))
	wu_error (BUG,"tree structure, ranges out in insert, amal node");

          
      if (range_l == range_h ||
	  in_routerlist (color, tree->color) ||
	  in_iplist (rep, tree->reps)) {	
/* 	wu_error (DEBUG, "inserted %s at leaf (%d to %d)", */
/* 		  wu_router_string (color), */
/* 		  range_l, */
/* 		  range_h); */
/* 	tree->reps = ips_insert (rep, tree->reps); */
	tree->color = routers_insert_list (color, tree->color);
      }
      else { /* subdividing range */
	split = split_range (range_l, range_h);

	listptr = tree->reps;
	while (listptr != NULL) { /* start insert while loop */
	  foo = listptr->ip;

	  if (foo < split) {
	    tree->left = _wu_insert (range_l, split - 1, foo, tree->color, tree->left); 	  }
	  else {
	    tree->right = _wu_insert (split, range_h, foo, tree->color, tree->right); 	  }	

	  listptr=listptr->next;
	} /* end insert while loop */
	
	tree->flag = FLAG_OPEN;
	free (tree->reps);

	if (rep < split) {		/* rep in range_l */
	  tree->left = _wu_insert (range_l, split - 1, rep, color, tree->left); 
	}
	else {			/* rep in range_h */
	  tree->right = _wu_insert (split, range_h, rep, color, tree->right); 
	}
      				/* default trees */
      if (tree->left == NULL ) {
	new = (WU_NODE *) wmalloc (sizeof (WU_NODE));
	new->left = NULL;
	new->right = NULL;
	new->flag = FLAG_UNREP;
	new->color = NULL;
	new->range_l = range_l;
	new->range_h = split - 1;
	new->reps = NULL;
	tree->left = new;
      }
      if (tree->right == NULL) {
	new = (WU_NODE *) wmalloc (sizeof (WU_NODE));
	new->left = NULL;
	new->right = NULL;
	new->flag = FLAG_UNREP;
	new->color = NULL;
	new->range_l = split;
	new->range_h = range_h;
	new->reps = NULL;
	tree->right = new;
      }
      }	/* subdividing */
    } /* amalgamated */
  
    else { wu_error (BUG,"tree insert: unknown flag"); }
    return (tree);
  }
}


WU_NODE * wu_insert  (
		  IP range_l,	/* range INCLUDE range_l */
		  IP range_h,	/* range INCLUDES range_h */
		  IP rep,	/* rep is HOST order */
		  struct wu_reply * color, 
		  WU_NODE * tree) {
  struct routerlist * single;

  single = (struct routerlist *) wmalloc (sizeof (struct routerlist));

  single->next = NULL;
  single->reply = color;

  return _wu_insert (range_l,
		     range_h,
		     rep,
		     single,
		     tree);
}
