#include "wu_head.h"
#include "wu_error.h"


/*
 * insert in order an IP into the head of a list
 */
struct iplist * ips_insert (IP rep, struct iplist * foo) {
  struct iplist * new;

  if (foo == NULL) {
    new = (struct iplist *) wmalloc (sizeof (struct iplist));
    new->ip = rep;
    new->next = NULL;
    return (new);
  }
  else {
    if (rep == foo->ip) { 
      return foo;
    }
    else if (ip_cmp (rep, foo->ip)) {
      new = (struct iplist *) wmalloc (sizeof (struct iplist));
      new->ip = rep;
      new->next = foo;
      return (new);
    }
    else {
      foo->next = ips_insert (rep, foo->next);
      return (foo);
    }
  }
}


/*
 * insert a list into a list
 */
struct iplist * ips_insert_list (struct iplist * a, struct iplist * b) {
  struct iplist * listptr;

  if (a == NULL && b == NULL)
    return NULL;

  if (b == NULL) 
    return a;

  if (a == NULL)
    return b;

  listptr = a;
  while (listptr != NULL) {
    b = ips_insert (listptr->ip, b);
    listptr = listptr->next;
  }
  return b;
  //free(a);
}


struct netlist * nets_insert 
(IP l_ho, IP h_ho, struct netlist * foo) {
  struct netlist * new;

  new = (struct netlist *) wmalloc (sizeof (struct netlist));
  new->host = htonl(l_ho);
  new->mask = maskbits_to_int (h_ho - l_ho);
  new->next = NULL;

  if (new->mask > WU_IPVX_MAXSIZE || new->mask < 0)
    wu_error(BUG, "nets_insert: got mask of %d", new->mask);

  if (foo == NULL) {		/* at end of list */
    return new;
  }

  else if (new->host == foo->host && new->mask == foo->mask) {
    /* lists are same */
    free (new);
    return foo;
  }
  
  else {
    /* recurse */
    free (new);
    foo->next = nets_insert (l_ho, h_ho, foo->next);
    return foo;
  }
}


struct netlist * nets_insert_list (struct netlist * a, struct netlist * b) {
  struct netlist * listptr;
  IP tmp_lo;
  IP tmp_hi;

  if (a == NULL && b == NULL)
    return NULL;

  if (b == NULL) 
    return a;

  if (a == NULL)
    return b;

  listptr = a;
  while (listptr != NULL) {
    tmp_lo = ntohl (listptr->host);
    if (listptr->mask > WU_IPVX_MAXSIZE || listptr->mask < 0)
      wu_error (DEBUG_MORE, "nets_insert_list: mask %d", listptr->mask);
    tmp_hi = tmp_lo + int_to_maskbits(listptr->mask);
    b = nets_insert (tmp_lo, tmp_hi, b);
    listptr = listptr->next;
  }

  //free(a);
  return b;
}



int router_cmp (struct wu_reply * a, struct wu_reply * b) {
				/* timeouts! */
  if (replies_equal (a, b)) return 0;

  if (a->flag == WU_REPLY_TIMEOUT) {
    if (b->flag == WU_REPLY_TIMEOUT) return 0;
    else return -1;
  }
  else if (b->flag == WU_REPLY_TIMEOUT) {
    //a not a timeout
    return 1;
  }

  //  if (a->flag == WU_REPLY_TIMEOUT) return 1;
  //  if (b->flag == WU_REPLY_TIMEOUT) return 0;
  //  if (a->flag < b->flag) return 0;
  //  if (a->flag != WU_REPLY_TIMEOUT && b->flag != a->flag) {
  if (ntohl (a->respondent.ip_no) < ntohl(b->respondent.ip_no)) return -1;
  else if (ntohl (a->respondent.ip_no) > ntohl(b->respondent.ip_no)) return 1;
  //  else return 0;
  //  }
  //UNUSED:
  if (a->flag < b->flag)    return -1;
  else if (a->flag > b->flag) return 1;

  if (a->packettype < b->packettype)  return -1; 
  else if (a->packettype > b->packettype) return 1;
  
  if (a->packetcode < b->packetcode) return -1; 
  else if (a->packetcode > b->packetcode) return 1; 
    
  return 1;
}

  
struct routerlist * routers_insert
(struct wu_reply * r, struct routerlist * foo) {
  struct routerlist * new;
  int cmp;

  //  wu_error(DEBUG_MORE,"insert router: %s", wu_router_string(r));

  if (foo == NULL) { 
    new = (struct routerlist *) wmalloc (sizeof (struct routerlist));
    new->reply = r;
    new->next = NULL;
    return (new);
  }

  else {
    if ((cmp = router_cmp (r, foo->reply)) < 0) {
      new = (struct routerlist *) wmalloc (sizeof (struct routerlist));
      new->reply = r;
      new->next = foo;
      return (new);
    }
    else if (cmp > 0) {
      foo->next = routers_insert (r, foo->next);
      return (foo);
    }
    else if (cmp == 0) { return foo; }
  }
}


struct routerlist * routers_insert_list (struct routerlist * a, struct routerlist * b) {
  struct routerlist * listptr;
  
  if (a == NULL && b == NULL)
    return NULL;

  if (b == NULL) 
    return a;

  if (a == NULL)
    return b;

  listptr = a;
  while (listptr != NULL) {
    b = routers_insert (listptr->reply, b);
    listptr = listptr->next;
  }
  
  return b;
  //free(a);
}


struct wu_net * ordered_net_cat (struct wu_net * first, struct wu_net * second) {
  /*
   * put second at end of first, returns head
   */
  struct wu_net * ptr;
  if (first == NULL) return second;
  if (second == NULL) return first;

  ptr = first;
  while (ptr->next != NULL) 
    ptr = ptr->next;

  ptr->next = second;
  return first;
}
