#ifndef WU_HEAD_H
#define WU_HEAD_H

#include <stdio.h>
#include <libnet.h>
#include "wu_error.h"
#include <pcap.h>
#include <sys/types.h>


				/*
				 * DEFINES
				 */

#define LIBNET_ERR_BUF 100

#define WEX_NO -1		/* exit failure */
#define WEX_OK 0		/* exit ok */

#define WU_BIG_STRING 10000
#define WU_SMALL_STRING 42
#define WU_MAX_IPS_ON_A_SUBNET 100
#define WU_MAX_PACKET_SIZE 256	
#define WU_SNAPLEN 256
#define PCAP_READ_TIMEOUT 20
#define WU_DEFAULT_SIM_MASK WU_IPVX_MAXSIZE
#define WU_NONCONTIG 1
#define WU_ADJACENT 2
#define WU_ECMP 4
#define WU_IPVX_MAXSIZE 32	/* IPV4 */
#define WU_SPLIT_MODE 0x1
#define WU_CLOUD_MODE 0x2
#define WU_MODE_CONSTANT 0x4
#define WU_MODE_INCREASING 0x8
#define WU_MODE_FORCE_BASE 0x10
#define WU_MODE_RANDOM 0x20
#define WU_NOTRACE 0x40

#define CARE_ABOUT_PACKET_TYPE 0x66	
#define ONLY_CARE_ABOUT_IP 0x67

				/*
				 * END DEFINES
				 */



				/*
				 * STRUCTURES
				 */

typedef u_long IP;
				

typedef struct ip_lookupable {
  IP ip_no;
  char lookedup;
  char hostname[WU_SMALL_STRING + 1];
} IP2;


struct wu_control
{
  IP base_host;                  
  int base_mask;			
  IP cur_host;
  int cur_mask;
  struct _distribution_tree * targets; 

  u_short current_ttl;           
#define WU_DEF_CURRE 1
  int maxdepth;			 
#define WU_DEF_MAXDE 1
  int timeout_depth;		 /* timeouts after to stop scan */
#define WU_DEF_TIMEO 1

  /* network options: */
  int opt_parprobes;		   /* number of concurrent probes */
#define WU_DEF_OPT_P 1
  struct timeval opt_reading_wait;
#define WU_DEF_OPT_R_SEC 0
#define WU_DEF_OPT_R_USEC 200000
  int opt_probe_cnt;		   /* redundancy of probes */
#define WU_DEF_PROBE 1
  int protocol;                    /* transport protocol in use */
#define WU_DEF_PROTO IPPROTO_UDP
  u_char *device;                  /* device to use */
#define WU_DEF_DEVIC "eth0"
  pcap_t *pcap;                    /* pcap packet listening structure */
  struct sockaddr_in sin;          /* socket address structure */
  int sock;			   /* socket */
  int linkoff;                     /* linklayer size */

  char ipdumpmode;		   
  struct _distribution_tree * routers;   /* stores IPs seen */

  int sim_mask;			/* mask to use when comparing router ips */
#define WU_DEF_SIM_M 24
  char packtype;		/* compare ICMP packet types ? */
#define WU_DEF_PACKT ONLY_CARE_ABOUT_IP

  int mode;
#define WU_DEF_MODE (WU_SPLIT_MODE | WU_MODE_CONSTANT | WU_MODE_RANDOM)
  int scan_density;             /* mostly not used */
#define WU_DEF_SCAN_ 1
  int subnet_resolution_mask;   /* initial size of subnet to scan */
#define WU_DEF_SUBNE 24		

  /*misc: */
  int _subnet_extra_depth;	/* solves bug */
};


/*
 * Used during network probes.
 */
struct wu_scan {
  IP target;
#define WU_MAX_QUERIES 6
  int flag;
#define WU_NOT_PROBED -2
#define WU_SENT_MULTI -4
#define WU_GOT_MULTI 1
#define WU_MULTI_DONE 2
  u_char * packet;
  struct timeval timesent;
  struct wu_reply * replies[WU_MAX_QUERIES + 1];
  int numreplies;		/* of different replies */
};


/*
 * For analysis of completed probes
 */
typedef struct _list {
  IP this;
  struct _list * next;
} LIST;


/*
 * Main Wu algorithm structure used to guess routing tables:
 */
typedef struct _node {
  struct routerlist * color;
  int flag;			
#define FLAG_OPEN 1
#define FLAG_AMAL 2
#define FLAG_UNREP 4
  IP range_l;			/* HOST order */
  IP range_h;			/* HOST order */
  struct iplist * reps;
  struct _node * left;
  struct _node * right;
} WU_NODE;


/*
 * Characterizes replies to probes.
 */
struct wu_reply {
  int pingtime;			/* round trip time in MS */
  IP2 respondent;		
  int flag;			
#define WU_NO_REPLY -0x10		
#define WU_REPLY_TIMEOUT WU_NO_REPLY
#define WU_REPLY 1		/* reply recieved from probe */
#define WU_TTL_PROBLEM 2	/* (see traceroute man page) */
#define WU_ICMP_PROTOCOL_VIOLATION 0x4 
#define WU_FARM 0x8		/* flag for farm detection algorithm */
  int proto;			/* protocol of reply */
  int packettype;	
  int packetcode;	
  int farm_mask;		/* if WU_FARM is set, the subnet of farm */
  char pingflag;		
#define PING_REPLY 1		/* if src IP == dst IP of probe */
#define NO_PING 0
};


struct state_vars {
  int state;
  char timeouts;
};


/*
 * netlist, routerlist, iplist are straightforward linked list structures.
 * wu_net is similar.
 */
struct netlist {
  IP host;			/* net order */
  int mask;			/* net order */
  struct netlist * next;
};


struct routerlist {
  struct wu_reply * reply;
  struct routerlist * next;
};


struct iplist {
  IP ip;			/* order USUALLY HOST if from a node2list */
  struct iplist * next;
};


struct wu_net {
  struct netlist * nets; 
  struct routerlist * routers;
  struct iplist * reps;
  struct wu_net * next;
};

				/*
				 * END STRUCTURES
				 */



				/*
				 * PROTOTYPES
				 */

/* random */
int  get_random_bytes (void *buf, int numbytes);
int  get_random_int ();
short  get_random_ushort ();
struct  wu_reply  * get_random_color (int maskin);

/* ipv4 */
IP int_to_maskbits (int mask);
int maskbits_to_int (IP dist);
IP  split_range (IP l, IP h);
int ip_cmp (IP a, IP b);

/* stack for DFS */
void  init_stack ();
int  push (int thing);
int  pop ();
void  print_stack ();

/* DFS functions */
void  wu_dfs ( int depth);
struct  wu_net  * scan_multi (struct wu_net * toscan);
struct  wu_control  * inc_controls (struct wu_net * net);
struct  wu_control  * dec_controls ();

/* output */
char  * wu_net_string (IP ip, int mask);
char  * wu_router_string (struct wu_reply * router);
char  * wu_nets_string (struct netlist * nets);
char  * wu_ips_string (struct iplist * ips);
char  * wu_routers_string (struct routerlist * routers);
int  netwidth (struct netlist * nets);

/* main() functions */
WU_NODE * wu_scan ();
int wu_shutdown();
int wu_init_control();

/* probe selection */
struct wu_scan ** update_probe_targets ( int * numdone) ;
struct wu_scan ** init_random_distribution ( int * numdone) ;
int create_random_distribution (int num_ips, IP base_ip_no, int mask,  int subnet_jump,  struct wu_scan ** targets  ) ;

/* network primitives */
int wu_packet_build_udp( struct wu_scan * scan);
int wu_packet_probe( struct wu_scan * scan);
int wu_close_network ();

/* analysis functions */
WU_NODE * wu_insert (IP range_l, IP range_h, IP rep, struct wu_reply * color, WU_NODE * tree) ;
WU_NODE * wu_insert2 (IP, IP, IP, IP, WU_NODE *);
struct wu_net * wu_adjacent_heuristic (struct wu_net * in,  int mask, int replytype);
struct wu_net * wu_ecmp_heuristic (struct wu_net * in,  int mask, int replytype) ;
void  find_farms(WU_NODE *);
int contiguous_farm_heuristic (struct wu_reply * a, struct wu_reply * b);
struct wu_net * re_aggregate ( struct wu_net * messed ) ;
void  amalgamate (WU_NODE * tree);

/* memory management */
void  * wmalloc (size_t size);
struct wu_net * wipe_wu_net ( struct wu_net * in) ;
struct wu_reply * wipe_reply ( struct wu_reply * in) ;
struct iplist * wipe_iplist (struct iplist * in) ;
struct netlist * wipe_netlist (struct netlist * in) ;
struct routerlist * wipe_routerlist (struct routerlist * in) ;
WU_NODE * free_only_nodes ( WU_NODE * in) ; 

/* lookupable IP structure management */
IP2 new_ip2 (IP in) ;
char * wu_lookup (IP2 * ip) ;

/* list primitives */
struct  wu_net  * node2list (WU_NODE * node, struct wu_net ** in);
struct  iplist  * ips_insert (IP rep, struct iplist * foo);
struct  iplist  * ips_insert_list (struct iplist * a, struct iplist * b);
struct  netlist  * nets_insert_list (struct netlist * a, struct netlist * b);
struct  routerlist  * routers_insert_list (struct routerlist * a, struct routerlist * b);
int in_routerlist (struct routerlist * foo, struct routerlist * list);
int in_iplist (IP  foo, struct iplist * list) ;
void wu_update_replies (struct wu_scan * scan, struct wu_reply * reply);

/* debug functions */
void  debug_wu_net (struct wu_net * sf);
void  traverse_wu_net (struct wu_net * sf);
struct  wu_net  * ordered_net_cat (struct wu_net * first, struct wu_net * second);

				/*
				 * END PROTOTYPES
				 */




				/* 
				 * GLOBAL INSTANCE DEFINED IN MAIN 
				 */
#ifdef WU_MAIN
#define G 
#else 
#define G extern
#endif
G struct wu_control CNTRL;
#undef G

#endif 
